"""ELT Context."""

from __future__ import annotations

from collections import namedtuple
from typing import Any

from sqlalchemy.orm import Session

from meltano.core.job import Job
from meltano.core.logging.output_logger import OutputLogger
from meltano.core.plugin import PluginRef, PluginType
from meltano.core.plugin.error import PluginNotFoundError
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.plugin_invoker import PluginInvoker, invoker_factory
from meltano.core.project import Project
from meltano.core.project_plugins_service import ProjectPluginsService


class PluginContext(
    namedtuple("PluginContext", "plugin settings_service session")  # noqa: WPS606
):
    """Plugin Context container."""

    def __getattr__(self, attr: str) -> Any:
        """Get plugin attribute.

        Args:
            attr: Attribute name.

        Returns:
            Attribute value.
        """
        return getattr(self.plugin, attr)

    def get_config(self, name: str, **kwargs: Any) -> Any:
        """Get plugin config by name.

        Args:
            name: Setting name to retrieve config for.
            kwargs: Keyword arguments to pass to SettingService.

        Returns:
            Setting value, if found, else None.
        """
        return self.settings_service.get(name, session=self.session, **kwargs)

    def config_dict(self, **kwargs) -> dict:
        """Get plugins config dict.

        Args:
            kwargs: Keyword arguments to pass to SettingService.

        Returns:
            Plugins config dict.
        """
        return self.settings_service.as_dict(session=self.session, **kwargs)

    def config_env(self, **kwargs) -> dict[str, str]:
        """Get plugins config environment.

        Args:
            kwargs: Keyword arguments to pass to SettingService.

        Returns:
            Plugins config environment dict.
        """
        return self.settings_service.as_env(session=self.session, **kwargs)

    @property
    def env(self) -> dict:
        """Get complete plugin environment dict.

        Returns:
            Complete plugin environment dict.
        """
        return {**self.plugin.info_env, **self.config_env()}


class ELTContext:  # noqa: WPS230
    """ELT Context."""

    def __init__(
        self,
        project,
        job: Job | None = None,
        session=None,
        extractor: PluginContext | None = None,
        loader: PluginContext | None = None,
        transform: PluginContext | None = None,
        transformer: PluginContext | None = None,
        only_transform: bool | None = False,
        dry_run: bool | None = False,
        full_refresh: bool | None = False,
        select_filter: list | None = None,
        catalog: str | None = None,
        state: str | None = None,
        plugins_service: ProjectPluginsService | None = None,
        base_output_logger: OutputLogger | None = None,
    ):
        """Initialise ELT Context instance.

        Args:
            project: Meltano Project instance.
            job: Job instance.
            session: SQLAlchemy Session instance.
            extractor: Extractor to use.
            loader: Loader to use.
            transform: Transform to use.
            transformer: Transformer to use.
            only_transform: Flag. Only run transform.
            dry_run: Flag. Don't actually run.
            full_refresh: Flag. Ignore previous captured state.
            select_filter: Select filters to apply to extractor.
            catalog: Catalog to pass to extractor.
            state: State to pass to extractor.
            plugins_service: PluginsService to use.
            base_output_logger: OutputLogger to use.
        """
        self.project = project
        self.job = job
        self.session = session

        self.extractor = extractor
        self.loader = loader
        self.transform = transform
        self.transformer = transformer

        self.only_transform = only_transform
        self.dry_run = dry_run
        self.full_refresh = full_refresh
        self.select_filter = select_filter or []
        self.catalog = catalog
        self.state = state

        self.plugins_service = plugins_service or ProjectPluginsService(project)
        self.base_output_logger = base_output_logger

    @property
    def elt_run_dir(self):
        """Get the ELT run directory.

        Returns:
            The job dir, if a Job is provided, else None.
        """
        if self.job:
            return self.project.job_dir(self.job.job_name, str(self.job.run_id))

    def invoker_for(self, plugin_type: PluginType) -> PluginInvoker:
        """Get invoker for given plugin type.

        Args:
            plugin_type: Plugin type to get invoker for.

        Returns:
            A PluginInvoker.
        """
        plugin_contexts = {
            PluginType.EXTRACTORS: self.extractor,
            PluginType.LOADERS: self.loader,
            PluginType.TRANSFORMERS: self.transformer,
        }

        plugin_context = plugin_contexts[plugin_type]
        return invoker_factory(
            self.project,
            plugin_context.plugin,
            context=self,
            run_dir=self.elt_run_dir,
            plugins_service=self.plugins_service,
            plugin_settings_service=plugin_context.settings_service,
        )

    def extractor_invoker(self) -> PluginInvoker:
        """Get the extractors' invoker.

        Returns:
            Invoker for contexts transformer.
        """
        return self.invoker_for(PluginType.EXTRACTORS)

    def loader_invoker(self) -> PluginInvoker:
        """Get the loaders' invoker.

        Returns:
            Invoker for contexts loader.
        """
        return self.invoker_for(PluginType.LOADERS)

    def transformer_invoker(self) -> PluginInvoker:
        """Get the transformers' invoker.

        Returns:
            Invoker for contexts transformer.
        """
        return self.invoker_for(PluginType.TRANSFORMERS)


class ELTContextBuilder:  # noqa: WPS214
    """ELT Context Builder."""

    def __init__(
        self,
        project: Project,
        plugins_service: ProjectPluginsService | None = None,
    ):
        """Instantiate new ELTContextBuilder.

        Args:
            project: A Meltano Project instance.
            plugins_service: A PluginsService instance.
        """
        self.project = project
        self.plugins_service = plugins_service or ProjectPluginsService(project)

        self._session = None
        self._job = None

        self._extractor = None
        self._loader = None
        self._transform = None
        self._transformer = None

        self._only_transform = False
        self._dry_run = False
        self._full_refresh = False
        self._select_filter = None
        self._catalog = None
        self._state = None
        self._base_output_logger = None

    def with_session(self, session: Session) -> ELTContextBuilder:
        """Include session when building context.

        Args:
            session: A SQLAlchemy session.

        Returns:
            Updated ELTContextBuilder instance.
        """
        self._session = session
        return self

    def with_job(self, job: Job) -> ELTContextBuilder:
        """Include job when building context.

        Args:
            job: Job instance.

        Returns:
            Updated ELTContextBuilder instance.
        """
        self._job = job
        return self

    def with_extractor(self, extractor_name: str) -> ELTContextBuilder:
        """Include extractor when building context.

        Args:
            extractor_name: Extractor name.

        Returns:
            Updated ELTContextBuilder instance.
        """
        self._extractor = PluginRef(PluginType.EXTRACTORS, extractor_name)
        return self

    def with_loader(self, loader_name: str) -> ELTContextBuilder:
        """Include loader when building context.

        Args:
            loader_name: Loader name.

        Returns:
            Updated ELTContextBuilder instance.
        """
        self._loader = PluginRef(PluginType.LOADERS, loader_name)
        return self

    def with_transform(self, transform: str) -> ELTContextBuilder:
        """Include transform when building context.

        Args:
            transform: Choice of "skip", "only" or "run".

        Returns:
            Updated ELTContextBuilder instance.
        """
        if transform == "skip":
            return self

        if transform not in {"run", "only"}:
            self._transform = PluginRef(PluginType.TRANSFORMS, transform)

        return self.with_transformer("dbt")

    def with_transformer(self, transformer_name: str) -> ELTContextBuilder:
        """Include transformer when building context.

        Args:
            transformer_name: Name of transformer.

        Returns:
            Updated ELTContextBuilder instance.
        """
        self._transformer = PluginRef(PluginType.TRANSFORMERS, transformer_name)
        return self

    def with_only_transform(self, only_transform: bool) -> ELTContextBuilder:
        """Include only transform flag when building context.

        Args:
            only_transform: Flag. Only run transform.

        Returns:
            Updated ELTContextBuilder instance.
        """
        self._only_transform = only_transform
        return self

    def with_dry_run(self, dry_run: bool) -> ELTContextBuilder:
        """Include dry run flag when building context.

        Args:
            dry_run: Flag. Do not actually run.

        Returns:
            Updated ELTContextBuilder instance.
        """
        self._dry_run = dry_run
        return self

    def with_full_refresh(self, full_refresh: bool) -> ELTContextBuilder:
        """Include full refresh flag when building context.

        Args:
            full_refresh: Flag. Perform a full refresh (ignore state left behind by any previous runs).

        Returns:
            Updated ELTContextBuilder instance.
        """
        self._full_refresh = full_refresh
        return self

    def with_select_filter(self, select_filter: list[str]) -> ELTContextBuilder:
        """Include select filters when building context.

        Args:
            select_filter: A list of select filter strings.

        Returns:
            Updated ELTContextBuilder instance.
        """
        self._select_filter = select_filter
        return self

    def with_catalog(self, catalog: str) -> ELTContextBuilder:
        """Include catalog file path when building context.

        Args:
            catalog: Path to singer catalog file.

        Returns:
            Updated ELTContextBuilder instance.
        """
        self._catalog = catalog
        return self

    def with_state(self, state: str) -> ELTContextBuilder:
        """Include state file path when building context.

        Args:
            state: Path to a singer state file.

        Returns:
            Updated ELTContextBuilder instance.
        """
        self._state = state
        return self

    def set_base_output_logger(self, base_output_logger: OutputLogger):  # noqa: WPS615
        """Set the base output logger for use in this ELTContext.

        Args:
            base_output_logger: The OutputLogger to use.
        """
        self._base_output_logger = base_output_logger

    def plugin_context(
        self,
        plugin_ref: PluginRef,
        env: dict | None = None,
        config: dict | None = None,
    ) -> PluginContext:
        """Create context object for a plugin.

        Args:
            plugin_ref: Plugin reference object.
            env: Environment override dictionary.
            config: Plugin configuration override dictionary.

        Returns:
            A new `PluginContext` object.

        Raises:
            PluginNotFoundError: if a plugin specified cannot be found.
        """
        try:
            plugin = self.plugins_service.get_plugin(plugin_ref)
        except PluginNotFoundError as err:
            if plugin_ref.name == "dbt":
                raise PluginNotFoundError(
                    "Transformer 'dbt' not found.\n"
                    + "Use of the legacy 'dbt' Transformer is deprecated in favor of "
                    + "new adapter specific implementations (e.g. 'dbt-snowflake') "
                    + "compatible with the 'meltano run ...' command.\n"
                    + "https://docs.meltano.com/guide/transformation\n"
                    + "To continue using the legacy 'dbt' Transformer, "
                    + "add it to your Project using 'meltano add transformer dbt'."
                ) from err
            raise

        return PluginContext(
            plugin=plugin,
            settings_service=PluginSettingsService(
                self.project,
                plugin,
                plugins_service=self.plugins_service,
                env_override=env,
                config_override=config,
            ),
            session=self._session,
        )

    def context(self) -> ELTContext:
        """Return ELT context.

        Returns:
            ELTContext instance.
        """
        env = {}

        extractor = None
        if self._extractor:
            config = {}
            if self._select_filter:
                config["_select_filter"] = self._select_filter
            if self._catalog:
                config["_catalog"] = self._catalog
            if self._state:
                config["_state"] = self._state

            extractor = self.plugin_context(
                self._extractor,
                config=config,
            )

            env.update(extractor.env)

        loader = None
        if self._loader:
            loader = self.plugin_context(
                self._loader,
                env=env.copy(),
            )

            env.update(loader.env)

        transform = None
        if self._transform:
            transform = self.plugin_context(
                self._transform,
                env=env.copy(),
            )

            env.update(transform.env)

        transformer = None
        if self._transformer:
            transformer = self.plugin_context(
                self._transformer,
                env=env.copy(),
            )

        return ELTContext(
            self.project,
            job=self._job,
            session=self._session,
            extractor=extractor,
            loader=loader,
            transform=transform,
            transformer=transformer,
            only_transform=self._only_transform,
            dry_run=self._dry_run,
            full_refresh=self._full_refresh,
            select_filter=self._select_filter,
            catalog=self._catalog,
            state=self._state,
            plugins_service=self.plugins_service,
            base_output_logger=self._base_output_logger,
        )



"""Basic EL style BlockSet implementation."""

from __future__ import annotations

import asyncio
import logging
from contextlib import asynccontextmanager, closing
from pathlib import Path
from typing import AsyncIterator

import structlog
from sqlalchemy.orm import Session

from meltano.core.db import project_engine
from meltano.core.elt_context import PluginContext
from meltano.core.job import Job, JobFinder
from meltano.core.job.stale_job_failer import fail_stale_jobs
from meltano.core.job_state import STATE_ID_COMPONENT_DELIMITER
from meltano.core.logging import JobLoggingService, OutputLogger
from meltano.core.plugin import PluginType
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.plugin_invoker import PluginInvoker, invoker_factory
from meltano.core.project import Project
from meltano.core.project_plugins_service import ProjectPluginsService
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.runner import RunnerError
from meltano.core.state_service import StateService

from .blockset import BlockSet, BlockSetValidationError
from .future_utils import first_failed_future, handle_producer_line_length_limit_error
from .ioblock import IOBlock
from .singer import SingerBlock

logger = structlog.getLogger(__name__)


class BlockSetHasNoStateError(Exception):
    """Occurs when state_service is accessed for ExtractLoadBlocks instance for which no block has state."""


class ELBContext:  # noqa: WPS230
    """ELBContext holds the context for ELB BlockSets."""

    def __init__(
        self,
        project: Project,
        plugins_service: ProjectPluginsService | None = None,
        session: Session | None = None,
        job: Job | None = None,
        full_refresh: bool | None = False,
        force: bool | None = False,
        update_state: bool | None = True,
        state_id_suffix: str | None = None,
        base_output_logger: OutputLogger | None = None,
    ):
        """Use an ELBContext to pass information on to ExtractLoadBlocks.

        Args:
            project: The project to use.
            plugins_service: The plugins service to use.
            session: The session to use.
            job: The job within this context should run.
            full_refresh: Whether this is a full refresh.
            force: Whether to force the execution of the job if it is stale.
            update_state: Whether to update the state of the job.
            state_id_suffix: The state ID suffix to use.
            base_output_logger: The base logger to use.
        """
        self.project = project
        self.plugins_service = plugins_service or ProjectPluginsService(project)

        self.session = session
        self.job = job

        self.full_refresh = full_refresh
        self.force = force
        self.update_state = update_state
        self.state_id_suffix = state_id_suffix

        # not yet used but required to satisfy the interface
        self.dry_run = False
        self.state = None
        self.select_filter = []
        self.catalog = None

        self.base_output_logger = base_output_logger

    @property
    def elt_run_dir(self) -> Path | None:
        """Obtain the run directory for the current job.

        Returns:
            The run directory for the current job.
        """
        if self.job:
            return self.project.job_dir(self.job.job_name, str(self.job.run_id))


class ELBContextBuilder:  # noqa: WPS214
    """Build up ELBContexts for ExtractLoadBlocks."""

    def __init__(
        self,
        project: Project,
        plugins_service: ProjectPluginsService | None = None,
        job: Job | None = None,
    ):
        """Initialize a new `ELBContextBuilder` that can be used to upgrade plugins to blocks for use in a ExtractLoadBlock.

        Args:
            project: the meltano project for the context.
            plugins_service: the plugins service for the context.
            job: the job for this ELT run.
        """
        self.project = project
        self.plugins_service = plugins_service or ProjectPluginsService(project)

        _, session_maker = project_engine(project)
        self.session = session_maker()

        self._job = None
        self._full_refresh = False
        self._state_update = True
        self._force = False
        self._state_id_suffix = None
        self._env = {}
        self._blocks = []

        self._base_output_logger = None

    def with_job(self, job: Job):
        """Set the associated job for the context.

        Args:
            job: the initial job for the context.

        Returns:
            self
        """
        self._job = job
        return self

    def with_full_refresh(self, full_refresh: bool):
        """Set whether this is a full refresh.

        Args:
            full_refresh : whether this is a full refresh.

        Returns:
            self
        """
        self._full_refresh = full_refresh
        return self

    def with_no_state_update(self, no_state_update: bool):
        """Set whether this run should not update state.

        By default we typically attempt to track state. This allows avoiding state management.


        Args:
            no_state_update: whether this run should update state.

        Returns:
            self
        """
        self._state_update = not no_state_update
        return self

    def with_force(self, force: bool):
        """Set whether the execution of the job should be forced if it is stale.

        Args:
            force: Whether to force the execution of the job if it is stale.

        Returns:
            self
        """
        self._force = force
        return self

    def with_state_id_suffix(self, state_id_suffix: str):
        """Set a state ID suffix for this run.

        Args:
            state_id_suffix: The suffix value.

        Returns:
            self
        """
        self._state_id_suffix = state_id_suffix
        return self

    def make_block(
        self,
        plugin: ProjectPlugin,
        plugin_args: list[str] | None = None,
    ) -> SingerBlock:
        """Create a new `SingerBlock` object, from a plugin.

        Args:
            plugin: The plugin to be executed.
            plugin_args: The arguments to be passed to the plugin.

        Returns:
            The new `SingerBlock` object.
        """
        ctx = self.plugin_context(plugin, env=self._env.copy())

        block = SingerBlock(
            block_ctx=ctx,
            project=self.project,
            plugins_service=self.plugins_service,
            plugin_invoker=self.invoker_for(ctx),
            plugin_args=plugin_args,
        )
        self._blocks.append(block)
        self._env.update(ctx.env)
        return block

    def plugin_context(
        self,
        plugin: ProjectPlugin,
        env: dict | None = None,
    ) -> PluginContext:
        """Create context object for a plugin.

        Args:
            plugin: The plugin to create the context for.
            env: Environment override dictionary.

        Returns:
            A new `PluginContext` object.
        """
        return PluginContext(
            plugin=plugin,
            settings_service=PluginSettingsService(
                self.project,
                plugin,
                plugins_service=self.plugins_service,
                env_override=env,
            ),
            session=self.session,
        )

    def invoker_for(
        self,
        plugin_context: PluginContext,
    ) -> PluginInvoker:
        """Create an invoker for a plugin from a PluginContext.

        Args:
            plugin_context: The plugin context to pass to the invoker_factory.

        Returns:
            A new `PluginInvoker` object.
        """
        return invoker_factory(
            self.project,
            plugin_context.plugin,
            context=self.context(),
            run_dir=self.elt_run_dir,
            plugins_service=self.plugins_service,
            plugin_settings_service=plugin_context.settings_service,
        )

    @property
    def elt_run_dir(self) -> Path | None:
        """Get the run directory for the current job.

        Returns:
            The run directory for the current job.
        """
        if self._job:
            return self.project.job_dir(self._job.job_name, str(self._job.run_id))

    def context(self) -> ELBContext:
        """Create an ELBContext object from the current builder state.

        Returns:
            A new `ELBContext` object.
        """
        return ELBContext(
            project=self.project,
            plugins_service=self.plugins_service,
            session=self.session,
            job=self._job,
            full_refresh=self._full_refresh,
            force=self._force,
            update_state=self._state_update,
            state_id_suffix=self._state_id_suffix,
            base_output_logger=self._base_output_logger,
        )


class ExtractLoadBlocks(BlockSet):  # noqa: WPS214
    """A basic BlockSet interface implementation that supports running basic EL (extract, load) patterns."""

    def __init__(
        self,
        context: ELBContext,
        blocks: tuple[IOBlock],
    ):
        """Initialize a basic BlockSet suitable for executing ELT tasks.

        Args:
            context: the elt context to use for this elt run.
            blocks: the IOBlocks that should be used for this elt run.
        """
        self.context = context
        self.blocks = blocks
        self.project_settings_service = ProjectSettingsService(
            self.context.project,
            config_service=self.context.plugins_service.config_service,
        )

        self.output_logger = OutputLogger(None)

        if not self.context.project.active_environment:
            self.context.job = None

        elif self.context.update_state:
            state_id = generate_state_id(
                self.context.project, self.context.state_id_suffix, self.head, self.tail
            )
            self.context.job = Job(job_name=state_id)
            job_logging_service = JobLoggingService(self.context.project)
            log_file = job_logging_service.generate_log_name(
                self.context.job.job_name, self.context.job.run_id
            )
            self.output_logger = OutputLogger(log_file)

        self._process_futures = None
        self._stdout_futures = None
        self._stderr_futures = None
        self._errors = []
        self._state_service = None

    def has_state(self) -> bool:
        """Check to see if any block in this BlockSet has 'state' capability.

        Returns:
            bool indicating whether BlockSet has 'state' capability
        """
        for block in self.blocks:
            if block.has_state:
                return True
        return False

    @property
    def state_service(self) -> StateService:
        """Get StateService for managing state for this BlockSet.

        Returns:
            A StateService using this BlockSet's context's session

        Raises:
            BlockSetHasNoStateError: if no Block in this BlockSet has state capability
        """
        if not self._state_service:
            if self.has_state():
                self._state_service = StateService(
                    project=self.context.project, session=self.context.session
                )
            else:
                raise BlockSetHasNoStateError()
        return self._state_service

    def index_last_input_done(self) -> int | None:
        """Return index of the block furthest from the start that has exited and required input.

        Returns:
            The index of the block furthest from the start that has exited and required input.
        """
        for idx, block in reversed(list(enumerate(self.blocks))):
            if block.requires_input and block.proxy_stderr.done():
                return idx

    def upstream_complete(self, index: int) -> bool | None:
        """Return whether blocks upstream from a given block index are already done.

        Args:
            index: The index of the block to check upstream from.

        Returns:
            True if all upstream blocks are done, False otherwise.
        """
        for idx, block in enumerate(self.blocks):
            if idx >= index:
                return True
            if block.process_future.done():
                continue
            return False

    async def upstream_stop(self, index) -> None:
        """Stop all blocks upstream of a given index.

        Args:
            index: The index of the block to stop upstream from.
        """
        for block in reversed(self.blocks[:index]):
            await block.stop()

    async def process_wait(
        self, output_exception_future: asyncio.Task | None, subset: int | None = None
    ) -> set[asyncio.Task]:
        """Wait on all process futures in the block set.

        Args:
            output_exception_future: additional future to wait on for output exceptions.
            subset: the subset of blocks to wait on.

        Returns:
            The set of all process futures + optional output exception futures.
        """
        futures = list(self.process_futures[subset:])
        if output_exception_future:
            futures.append(output_exception_future)

        done, _ = await asyncio.wait(
            futures,
            return_when=asyncio.FIRST_COMPLETED,
        )
        return done

    def validate_set(self) -> None:  # noqa: WPS231, WPS238
        """Validate a ExtractLoad block set to ensure its valid and runnable.

        Raises:
            BlockSetValidationError: if the block set is not valid.
        """
        if not self.blocks:
            raise BlockSetValidationError("No blocks in set.")

        if self.head.consumer:
            raise BlockSetValidationError("first block in set should not be consumer")

        if self.tail.producer:
            raise BlockSetValidationError("last block in set should not be a producer")

        for block in self.intermediate:
            if not block.consumer or not block.producer:
                raise BlockSetValidationError(
                    "intermediate blocks must be producers AND consumers"
                )

    async def execute(self) -> None:
        """Build the IO chain and execute the actual ELT task."""
        async with self._start_blocks():
            await self._link_io()
            manager = ELBExecutionManager(self)
            await manager.run()

    async def run(self) -> None:
        """Run the ELT task."""
        if self.context.job:
            # TODO: legacy `meltano elt` style logging should be deprecated
            legacy_log_handler = self.output_logger.out("meltano", logger)
            with legacy_log_handler.redirect_logging():
                await self.run_with_job()
                return
        else:
            logger.warning(
                "No active environment, proceeding with stateless run! See https://docs.meltano.com/reference/command-line-interface#run for details."
            )
        await self.execute()

    async def run_with_job(self) -> None:
        """Run the ELT task within the context of a job.

        Raises:
            RunnerError: if failures are encountered during execution or if the underlying pipeline/job is already running.
        """
        job = self.context.job
        fail_stale_jobs(self.context.session, job.job_name)
        if not self.context.force:
            existing = JobFinder(job.job_name).latest_running(self.context.session)
            if existing:
                raise RunnerError(
                    f"Another '{job.job_name}' pipeline is already running which started at {existing.started_at}. "
                    + "To ignore this check use the '--force' option."
                )

        with closing(self.context.session) as session:
            async with job.run(session):
                await self.execute()

    async def terminate(self, graceful: bool = False) -> None:
        """Terminate an in flight ExtractLoad execution, potentially disruptive.

        Not actually implemented yet.

        Args:
            graceful: Whether or not the BlockSet should try to gracefully quit.

        Raises:
            NotImplementedError: if graceful termination is not implemented.
        """
        if graceful:
            raise NotImplementedError

        for block in self.blocks:
            await block.stop(kill=True)

    @property
    def process_futures(self) -> list[asyncio.Task]:
        """Return the futures of the blocks subprocess calls.

        Returns:
            The list of all process futures.
        """
        if self._process_futures is None:
            self._process_futures = [block.process_future for block in self.blocks]
        return self._process_futures

    @property
    def stdout_futures(self) -> list[asyncio.Task]:
        """Access the futures of the blocks stdout proxy tasks.

        Returns:
            The list of all stdout futures.
        """
        if self._stdout_futures is None:
            self._stdout_futures = [block.proxy_stdout() for block in self.blocks]
        return self._stdout_futures

    @property
    def stderr_futures(self) -> list[asyncio.Task]:
        """Access the futures of the blocks stderr proxy tasks.

        Returns:
            The list of all stderr futures.
        """
        if self._stderr_futures is None:
            self._stderr_futures = [block.proxy_stderr() for block in self.blocks]
        return self._stderr_futures

    @property
    def head(self) -> IOBlock:
        """Obtain the first block in the block set.

        Returns:
            The first block in the block set.
        """
        return self.blocks[0]

    @property
    def tail(self) -> IOBlock:
        """Obtain the last block in the block set.

        Returns:
            The last block in the block set.
        """
        return self.blocks[-1]

    @property
    def intermediate(self) -> tuple[IOBlock]:
        """Obtain the intermediate blocks in the set - excluding the first and last block.

        Returns:
            The intermediate blocks in the block set.
        """
        return self.blocks[1:-1]

    @asynccontextmanager
    async def _start_blocks(self) -> AsyncIterator[None]:
        """Start the blocks in the block set.

        Yields:
            None
        """
        try:  # noqa:  WPS229
            for block in self.blocks:
                await block.pre(self.context)
                await block.start()
            yield
        finally:
            await self._cleanup()

    async def _cleanup(self) -> None:
        for block in self.blocks:
            await block.post()

    async def _link_io(self) -> None:  # noqa: WPS231
        """Link the blocks in the set together.

        This method does one last validation check to ensure that a consumer has a producer upstream.

        Raises:
            BlockSetValidationError: if consumer does not have an upstream producer.
        """
        for idx, block in enumerate(self.blocks):
            logger_base = logger.bind(
                consumer=block.consumer,
                producer=block.producer,
                string_id=block.string_id,
                cmd_type="elb",
            )
            if logger.isEnabledFor(logging.DEBUG):
                block.stdout_link(
                    self.output_logger.out(
                        block.string_id,
                        logger_base.bind(stdio="stdout"),
                    )
                )
            block.stderr_link(
                self.output_logger.out(
                    block.string_id,
                    logger_base.bind(stdio="stderr"),
                )
            )
            if block.consumer:
                if idx != 0 and self.blocks[idx - 1].producer:
                    self.blocks[idx - 1].stdout_link(
                        block.stdin
                    )  # link previous blocks stdout with current blocks stdin
                else:
                    raise BlockSetValidationError(
                        "run step requires input but has no upstream"
                    )


class ELBExecutionManager:
    """Execution manager for ExtractLoadBlock sets."""

    def __init__(self, elb: ExtractLoadBlocks) -> None:
        """Initialize the ELBExecutionManager which will handle the actual runtime management of the block set.

        Args:
            elb: The ExtractLoadBlocks to manage.
        """
        self.elb = elb
        self.stream_buffer_size = self.elb.project_settings_service.get(
            "elt.buffer_size"
        )
        self.line_length_limit = self.stream_buffer_size // 2

        self._producer_code = None
        self._consumer_code = None
        self._intermediate_codes: dict[str, int] = {}

    async def run(self) -> None:
        """Run is used to actually perform the execution of the ExtractLoadBlock set.

        That entails starting the blocks, waiting for them to complete, ensuring that exceptions are handled, and
        stopping blocks or waiting for IO to complete as appropriate. Expect a RunnerError to be raised if any of
        the blocks exit with a non 0 exit code.
        """
        await self._wait_for_process_completion(self.elb.head)
        _check_exit_codes(
            self._producer_code, self._consumer_code, self._intermediate_codes
        )

    async def _complete_upstream(self) -> None:
        """Wait for the upstream blocks to complete."""
        producer = self.elb.head
        consumer = self.elb.tail

        if self.elb.upstream_complete(len(self.elb.blocks) - 1):
            self._producer_code = producer.process_future.result()
        else:
            # If the last consumer (target) completes before the upstream producers, it failed before processing all
            # output. So we should kill the upstream producers and cancel output processing since there's no final
            # destination to forward output to.
            await self.elb.upstream_stop(len(self.elb.blocks) - 1)
            # Pretend the producer (tap) finished successfully since it didn't itself fail
            self._producer_code = 0
        # Wait for all buffered consumer (target) output to be processed
        await asyncio.wait([consumer.proxy_stdout(), consumer.proxy_stderr()])

    async def _wait_for_process_completion(  # noqa: WPS213 WPS217 WPS210
        self, current_head: IOBlock
    ) -> tuple[int, int] | None:
        """Wait for the current head block to complete or for an error to occur.

        Args:
            current_head: The current head block

        Returns:
            A tuple of the producer exit code and the consumer exit code

        Raises:
            RunnerError: if any intermediate blocks failed.
            exception: if any of the output futures encountered an exception.
        """
        start_idx = self.elb.blocks.index(current_head)
        remaining_blocks = self.elb.blocks[start_idx:]

        if remaining_blocks is None or current_head == self.elb.tail:
            return

        stdout_futures = [block.proxy_stdout() for block in remaining_blocks]
        stderr_futures = [block.proxy_stderr() for block in remaining_blocks]

        output_exception_future = asyncio.ensure_future(
            asyncio.wait(
                [*stdout_futures, *stderr_futures],
                return_when=asyncio.FIRST_EXCEPTION,
            )
        )

        logger.debug("waiting for process completion or exception")
        done = await self.elb.process_wait(output_exception_future, start_idx)

        output_futures_failed = first_failed_future(output_exception_future, done)
        if output_futures_failed:
            # Special behavior for a producer stdout handler raising a line length limit error.
            if self.elb.head.proxy_stdout() == output_futures_failed:
                handle_producer_line_length_limit_error(
                    output_futures_failed.exception(),
                    line_length_limit=self.line_length_limit,
                    stream_buffer_size=self.stream_buffer_size,
                )
            raise output_futures_failed.exception()
        else:
            # If all the output handlers completed without raising an exception,
            # we still need to wait for all the underlying block processes to complete.
            # note that since all output handlers completed we DO NOT need to wait for any output futures!
            done = await self.elb.process_wait(None, start_idx)
        if self.elb.tail.process_future.done():
            logger.debug("tail consumer completed first")
            self._consumer_code = self.elb.tail.process_future.result()
            await self._complete_upstream()
        elif current_head.process_future.done():
            logger.debug(
                "head producer completed first as expected", name=current_head.string_id
            )
            await self._handle_head_completed(current_head, start_idx)
        else:
            logger.warning("Intermediate block in sequence failed.")
            await self._stop_all_blocks(start_idx)
            raise RunnerError(
                "Unexpected completion sequence in ExtractLoadBlock set. Intermediate block (likely a mapper) failed.",
                {
                    PluginType.EXTRACTORS: 1,
                    PluginType.LOADERS: 1,
                },
            )

    async def _handle_head_completed(
        self, current_head: IOBlock, start_idx: int
    ) -> None:
        next_head: IOBlock = self.elb.blocks[start_idx + 1]

        if current_head is self.elb.head:
            self._producer_code = current_head.process_future.result()
        else:
            self._intermediate_codes[
                current_head.string_id
            ] = current_head.process_future.result()

        await asyncio.wait([current_head.proxy_stdout(), current_head.proxy_stderr()])
        # Close next inline stdin so downstream can cascade and complete naturally
        await next_head.close_stdin()

        if next_head is self.elb.tail:
            logger.debug("tail consumer is next block, wrapping up")
            # Wait for all buffered target output to be processed
            await asyncio.wait([next_head.proxy_stdout(), next_head.proxy_stderr()])

            # Wait for target process future to complete
            await next_head.process_future
            self._consumer_code = next_head.process_future.result()
            return  # break our recursion

        await self._wait_for_process_completion(next_head)

    async def _stop_all_blocks(self, idx: int = 0) -> None:
        """Close stdin and stop all blocks inclusive of index.

        Args:
            idx: starting index of the block's to stop.
        """
        for block in self.elb.blocks[idx:]:
            await block.close_stdin()
            await block.stop()


def _check_exit_codes(  # noqa: WPS238
    producer_code: int, consumer_code: int, intermediate_codes: dict[str, int]
) -> None:
    """Check exit codes for failures, and raise the appropriate RunnerError if needed.

    Args:
        producer_code: exit code of the producer (tap)
        consumer_code: exit code of the consumer (target)
        intermediate_codes: exit codes of the intermediate blocks (mappers)

    Raises:
        RunnerError: if the producer, consumer, or mapper exit codes are non-zero
    """
    if producer_code and consumer_code:
        raise RunnerError(
            "Extractor and loader failed",
            {PluginType.EXTRACTORS: producer_code, PluginType.LOADERS: consumer_code},
        )

    if producer_code:
        raise RunnerError("Extractor failed", {PluginType.EXTRACTORS: producer_code})

    if consumer_code:
        raise RunnerError("Loader failed", {PluginType.LOADERS: consumer_code})

    failed_mappers = []
    for mapper_id in intermediate_codes.keys():
        if intermediate_codes[mapper_id]:
            failed_mappers.append({mapper_id: intermediate_codes[mapper_id]})

    if failed_mappers:
        raise RunnerError("Mappers failed", failed_mappers)


def generate_state_id(
    project: Project, state_id_suffix: str | None, consumer: IOBlock, producer: IOBlock
) -> str:
    """Generate a state id based on a project active environment and consumer and producer names.

    Args:
        project: Project to retrieve active environment from.
        state_id_suffix: State ID suffix value.
        consumer: Consumer block.
        producer: Producer block.

    Returns:
        State id string.

    Raises:
        RunnerError: if the project does not have an active environment.
    """
    if not project.active_environment:
        raise RunnerError(
            "No active environment for invocation, but requested state id"
        )

    state_id_components = [
        project.active_environment.name,
        f"{consumer.string_id}-to-{producer.string_id}",
        state_id_suffix or project.active_environment.state_id_suffix,
    ]

    if any(c for c in state_id_components if c and STATE_ID_COMPONENT_DELIMITER in c):
        raise RunnerError(
            f"Cannot generate a state ID from components containing the delimiter string '{STATE_ID_COMPONENT_DELIMITER}'"
        )

    return STATE_ID_COMPONENT_DELIMITER.join(c for c in state_id_components if c)



"""StateStoreManager for systemdb state backend."""
from __future__ import annotations

from sqlalchemy import select
from sqlalchemy.orm import Session

from meltano.core.job_state import JobState
from meltano.core.state_store.base import StateStoreManager
from meltano.core.utils import merge


class DBStateStoreManager(StateStoreManager):
    """StateStoreManager implementation for state stored in the system db."""

    label = "Database"

    def __init__(self, session: Session, **kwargs):
        """Initialize the DBStateStoreManager.

        Args:
            session: the db session to use
            kwargs: optional keyword args to supply to StateStoreManager
        """
        super().__init__(**kwargs)
        self.session = session

    def set(self, state: JobState) -> None:
        """Set the job state for the given state_id.

        Args:
            state: the state to set.
        """
        existing_job_state = (
            self.session.query(JobState)
            .filter(JobState.state_id == state.state_id)
            .first()
        )
        partial_state = state.partial_state
        completed_state = state.completed_state
        if existing_job_state:
            if existing_job_state.partial_state and not state.is_complete():
                partial_state = merge(
                    state.partial_state, existing_job_state.partial_state
                )
            if not state.is_complete():
                completed_state = existing_job_state.completed_state
        new_job_state = JobState(
            state_id=state.state_id,
            partial_state=partial_state,
            completed_state=completed_state,
        )
        if existing_job_state:
            self.session.delete(existing_job_state)
        self.session.add(new_job_state)
        self.session.commit()

    def get(self, state_id):
        """Get the job state for the given state_id.

        Args:
            state_id: the name of the job to get state for

        Returns:
            The current state for the given job
        """
        return (
            self.session.query(JobState).filter(JobState.state_id == state_id).first()
        )

    def clear(self, state_id):
        """Clear state for the given state_id.

        Args:
            state_id: the state_id to clear state for
        """
        job_state: JobState | None = (
            self.session.query(JobState).filter(JobState.state_id == state_id).first()
        )
        if job_state:
            self.session.delete(job_state)
            self.session.commit()

    def get_state_ids(self, pattern: str | None = None):
        """Get all state_ids available in this state store manager.

        Args:
            pattern: glob-style pattern to filter by

        Returns:
            Generator yielding names of available jobs
        """
        if pattern:
            return (
                job_state.state_id
                for job_state in self.session.query(JobState)
                .filter(JobState.state_id.like(pattern.replace("*", "%")))
                .all()
            )
        return (
            record[0]
            for record in self.session.execute(select(JobState.state_id)).all()
        )

    def acquire_lock(self, state_id):
        """Acquire a naive lock for the given job's state.

        For DBStateStoreManager, the db manages transactions.
        This does nothing.

        Args:
            state_id: the state_id to lock
        """
        ...  # noqa: WPS428

    def release_lock(self, state_id):
        """Release the lock for the given job's state.

        For DBStateStoreManager, the db manages transactions.
        This does nothing.

        Args:
            state_id: the state_id to unlock
        """
        ...  # noqa: WPS428



from __future__ import annotations

from secrets import token_hex

from meltano.api.config import ProjectSettings as APIProjectSettings

# session are used internally to verify the tokens
SECRET_KEY_SIZE = 64
SECRET_KEY = token_hex(SECRET_KEY_SIZE)


class ProjectSettings(APIProjectSettings):
    settings_map = {
        "SERVER_NAME": "ui.server_name",
        "SESSION_COOKIE_DOMAIN": "ui.session_cookie_domain",
        # Flask-Authlib
        "FACEBOOK_CLIENT_ID": "oauth_service.facebook.client_id",
        "FACEBOOK_CLIENT_SECRET": "oauth_service.facebook.client_secret",
        "GOOGLE_ADWORDS_CLIENT_ID": "oauth_service.google_adwords.client_id",
        "GOOGLE_ADWORDS_CLIENT_SECRET": "oauth_service.google_adwords.client_secret",
    }



"""Meltano Hub Client."""

from __future__ import annotations

from http import HTTPStatus
from typing import Any

import requests
from requests.adapters import HTTPAdapter
from structlog.stdlib import get_logger
from urllib3 import Retry

import meltano
from meltano.core.plugin import (
    BasePlugin,
    PluginDefinition,
    PluginRef,
    PluginType,
    Variant,
)
from meltano.core.plugin.error import PluginNotFoundError
from meltano.core.plugin.factory import base_plugin_factory
from meltano.core.plugin_discovery_service import PluginRepository
from meltano.core.project import Project
from meltano.core.project_settings_service import ProjectSettingsService

from .schema import IndexedPlugin, VariantRef

logger = get_logger(__name__)


class HubPluginTypeNotFoundError(Exception):
    """Raised when a Hub plugin type is not found."""

    def __init__(self, plugin_type: PluginType):
        """Create a new HubPluginVariantNotFound.

        Args:
            plugin_type: The type of the plugin.
        """
        self.plugin_type = plugin_type

    def __str__(self) -> str:
        """Return a string representation of the error.

        Returns:
            The string representation of the error.
        """
        return "{type} is not supported in Meltano Hub. Available plugin types: {types}".format(
            type=self.plugin_type.descriptor.capitalize(),
            types=PluginType.plurals(),
        )


class HubConnectionError(Exception):
    """Raised when a Hub connection error occurs."""

    def __init__(self, reason: str):
        """Create a new HubConnectionError.

        Args:
            reason: The reason for the error.
        """
        message = f"Could not connect to Meltano Hub. {reason}"
        super().__init__(message)


class HubPluginVariantNotFoundError(Exception):
    """Raised when a Hub plugin variant is not found."""

    def __init__(
        self,
        plugin_type: PluginType,
        plugin: IndexedPlugin,
        variant_name: str,
    ):
        """Create a new HubPluginVariantNotFound.

        Args:
            plugin_type: The type of the plugin.
            plugin: The indexed plugin.
            variant_name: The name of the variant that was not found.
        """
        self.plugin_type = plugin_type
        self.plugin = plugin
        self.variant_name = variant_name

    def __str__(self) -> str:
        """Return a string representation of the error.

        Returns:
            The string representation of the error.
        """
        return "{type} '{name}' variant '{variant}' is not known to Meltano. Variants: {variant_labels}".format(
            type=self.plugin_type.descriptor.capitalize(),
            name=self.plugin.name,
            variant=self.variant_name,
            variant_labels=self.plugin.variant_labels,
        )


class MeltanoHubService(PluginRepository):  # noqa: WPS214
    """PluginRepository implementation for the Meltano Hub."""

    def __init__(self, project: Project) -> None:
        """Initialize the service.

        Args:
            project: The Meltano project.
        """
        self.project = project
        self.session = requests.Session()
        self.session.headers.update(
            {
                "Accept": "application/json",
                "User-Agent": f"Meltano/{meltano.__version__}",
            }
        )

        self.settings_service = ProjectSettingsService(self.project)

        if self.settings_service.get("send_anonymous_usage_stats"):
            project_id = self.settings_service.get("project_id")

            self.session.headers["X-Project-ID"] = project_id

        if self.hub_url_auth:
            self.session.headers.update({"Authorization": self.hub_url_auth})

        adapter = HTTPAdapter(
            max_retries=Retry(
                total=3,
                backoff_factor=0,
                status_forcelist=[
                    HTTPStatus.TOO_MANY_REQUESTS,
                    HTTPStatus.INTERNAL_SERVER_ERROR,
                    HTTPStatus.BAD_GATEWAY,
                    HTTPStatus.SERVICE_UNAVAILABLE,
                    HTTPStatus.GATEWAY_TIMEOUT,
                ],
                raise_on_status=False,
            ),
        )
        self.session.mount("http://", adapter)
        self.session.mount("https://", adapter)

    @property
    def hub_api_url(self) -> str:
        """Return the URL of the Hub API.

        Returns:
            The URL of the Hub API.
        """
        hub_api_root = self.settings_service.get("hub_api_root")
        hub_url = self.settings_service.get("hub_url")

        return hub_api_root or f"{hub_url}/meltano/api/v1"

    @property
    def hub_url_auth(self):
        """Return the `hub_url_auth` setting.

        Returns:
            The `hub_url_auth` setting.
        """
        return self.settings_service.get("hub_url_auth")

    def plugin_type_endpoint(self, plugin_type: PluginType) -> str:
        """Return the list endpoint for the given plugin type.

        Args:
            plugin_type: The plugin type.

        Returns:
            The endpoint for the given plugin type.
        """
        return f"{self.hub_api_url}/plugins/{plugin_type.value}/index"

    def plugin_endpoint(
        self,
        plugin_type: PluginType,
        plugin_name: str,
        variant_name: str | None = None,
    ) -> str:
        """Return the resource endpoint for the given plugin.

        Args:
            plugin_type: The plugin type.
            plugin_name: The plugin name.
            variant_name: The plugin variant name.

        Returns:
            The endpoint for the given plugin type.
        """
        url = f"{self.hub_api_url}/plugins/{plugin_type.value}/{plugin_name}"
        if variant_name:
            url = f"{url}--{variant_name}"

        return url

    def find_definition(
        self,
        plugin_type: PluginType,
        plugin_name: str,
        variant_name: str | None = None,
    ) -> PluginDefinition:
        """Find a locked plugin definition.

        Args:
            plugin_type: The plugin type.
            plugin_name: The plugin name.
            variant_name: The plugin variant name.

        Returns:
            The plugin definition.

        Raises:
            PluginNotFoundError: If the plugin definition could not be found.
            HubPluginVariantNotFoundError: If the plugin variant could not be found.
            HubConnectionError: If the Hub API could not be reached.
        """
        plugins = self.get_plugins_of_type(plugin_type)

        try:
            plugin = plugins[plugin_name]
        except KeyError as plugins_key_err:
            raise PluginNotFoundError(
                PluginRef(plugin_type, plugin_name)
            ) from plugins_key_err

        if variant_name is None or variant_name in {
            Variant.DEFAULT_NAME,
            Variant.ORIGINAL_NAME,
        }:
            variant_name = plugin.default_variant

        try:
            url = plugin.variants[variant_name].ref
        except KeyError as variant_key_err:
            raise HubPluginVariantNotFoundError(
                plugin_type, plugin, variant_name
            ) from variant_key_err

        response = self.session.get(url)

        try:
            response.raise_for_status()
        except requests.HTTPError as http_err:
            logger.error(
                "Can not retrieve plugin",
                status_code=http_err.response.status_code,
                error=http_err,
            )
            raise HubConnectionError(str(http_err)) from http_err

        return PluginDefinition(**response.json(), plugin_type=plugin_type)

    def find_base_plugin(
        self,
        plugin_type: PluginType,
        plugin_name: str,
        variant: str | None = None,
    ) -> BasePlugin:
        """Get the base plugin for a project plugin.

        Args:
            plugin_type: The plugin type.
            plugin_name: The plugin name.
            variant: The plugin variant.

        Returns:
            The base plugin.
        """
        plugin = self.find_definition(
            plugin_type,
            plugin_name,
            variant_name=variant,
        )

        return base_plugin_factory(plugin, plugin.variants[0])

    def get_plugins_of_type(  # noqa: WPS210
        self,
        plugin_type: PluginType,
    ) -> dict[str, IndexedPlugin]:
        """Get all plugins of a given type.

        Args:
            plugin_type: The plugin type.

        Returns:
            The plugin definitions.

        Raises:
            HubPluginTypeNotFoundError: If the plugin type is not supported.
            HubConnectionError: If the Hub API could not be reached.
        """
        if not plugin_type.discoverable:
            return {}

        url = self.plugin_type_endpoint(plugin_type)
        response = self.session.get(url)

        try:
            response.raise_for_status()
        except requests.HTTPError as err:
            logger.error(
                "Can not retrieve plugin type",
                status_code=err.response.status_code,
                error=err,
            )
            if err.response.status_code < HTTPStatus.TOO_MANY_REQUESTS:
                raise HubPluginTypeNotFoundError(plugin_type) from err
            raise HubConnectionError(err.response.reason) from err

        plugins: dict[str, dict[str, Any]] = response.json()
        return {
            name: IndexedPlugin(
                name,
                logo_url=plugin["logo_url"],
                default_variant=plugin["default_variant"],
                variants={
                    variant_name: VariantRef(variant_name, ref=variant["ref"])
                    for variant_name, variant in plugin["variants"].items()
                },
            )
            for name, plugin in plugins.items()
        }



"""Response schemas for the Meltano Hub API."""

from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass, field


@dataclass
class IndexedPlugin:
    """Response schema for the plugin type index."""

    name: str
    default_variant: str
    variants: dict[str, VariantRef] = field(default_factory=dict)
    logo_url: str | None = None

    @property
    def variant_labels(self) -> list[str]:
        """Return a list of variant labels.

        Returns:
            A list of variant labels.
        """
        result = []
        variants = deepcopy(self.variants)
        default_variant = variants.pop(self.default_variant)
        result.append(f"{default_variant.name} (default)")

        for variant in variants.values():
            # Add here any other variant metadata like maintenance status, etc.
            result.append(variant.name)

        return result


@dataclass
class VariantRef:
    """Response schema for a plugin variant."""

    name: str
    ref: str



"""Legacy discovery file object."""

from __future__ import annotations

from meltano.core.behavior.canonical import Canonical
from meltano.core.plugin import PluginDefinition, PluginType


class DiscoveryFile(Canonical):
    """A discovery file object."""

    def __init__(self, version=1, **plugins):
        """Create a new DiscoveryFile.

        Args:
            version: The version of the discovery file.
            plugins: The plugins to add to the discovery file.
        """
        super().__init__(version=int(version))

        for ptype in PluginType:
            self[ptype] = []

        for plugin_type, raw_plugins in plugins.items():
            for raw_plugin in raw_plugins:
                plugin_def = PluginDefinition(
                    plugin_type,
                    raw_plugin.pop("name"),
                    raw_plugin.pop("namespace"),
                    **raw_plugin,
                )
                self[plugin_type].append(plugin_def)

    @classmethod
    def file_version(cls, attrs):
        """Return version of discovery file represented by attrs dictionary.

        Args:
            attrs: The attributes of the discovery file.

        Returns:
            The version of the discovery file.
        """
        return int(attrs.get("version", 1))



"""Defines PluginTestService."""

from __future__ import annotations

import asyncio
import json
import logging
from abc import ABC, abstractmethod

from meltano.core.plugin.base import PluginType
from meltano.core.plugin.error import PluginNotSupportedError
from meltano.core.plugin_invoker import PluginInvoker

logger = logging.getLogger(__name__)


class PluginTestServiceFactory:
    """Factory class to resolve a plugin test service."""

    def __init__(self, plugin_invoker: PluginInvoker):
        """Construct a PluginTestServiceFactory instance.

        Args:
            plugin_invoker: The invocation instance of the plugin to test.
        """
        self.plugin_invoker = plugin_invoker

    def get_test_service(self):
        """Resolve a test service instance for a plugin type.

        Returns:
            The test service instance.

        Raises:
            PluginNotSupportedError: If the plugin type is not supported for testing.
        """
        test_services = {PluginType.EXTRACTORS: ExtractorTestService}

        try:
            return test_services[self.plugin_invoker.plugin.type](self.plugin_invoker)
        except KeyError as err:
            raise PluginNotSupportedError(self.plugin_invoker.plugin) from err


class PluginTestService(ABC):
    """Abstract base class for plugin test operations."""

    def __init__(self, plugin_invoker: PluginInvoker):
        """Construct a PluginTestService instance.

        Args:
            plugin_invoker: The invocation instance of the plugin to test
        """
        self.plugin_invoker = plugin_invoker

    @abstractmethod
    async def validate(self) -> tuple[bool, str]:
        """Abstract method to validate plugin configuration."""


class ExtractorTestService(PluginTestService):
    """Handle extractor test operations."""

    async def validate(self) -> tuple[bool, str]:
        """Validate extractor configuration.

        Returns:
            The validation result and supporting context message (if applicable).
        """
        process = None

        try:
            process = await self.plugin_invoker.invoke_async(
                stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.STDOUT
            )
        except Exception as exc:
            return False, str(exc)

        last_line = None
        while not process.stdout.at_eof():
            data = await process.stdout.readline()
            line = data.decode("ascii").strip()
            if line:
                logger.debug(line)
                last_line = line

            try:
                message_type = json.loads(line)["type"]
            except (json.decoder.JSONDecodeError, KeyError):
                continue

            if message_type == "RECORD":
                process.terminate()
                break

        returncode = await process.wait()

        # considered valid if subprocess is terminated (exit status < 0) on RECORD message received
        # see https://docs.python.org/3/library/subprocess.html#subprocess.CompletedProcess.returncode
        return (
            returncode < 0,
            last_line if returncode else "No RECORD message received",
        )



"""Discoverable Plugins CLI."""

from __future__ import annotations

import click

from meltano.core.hub import MeltanoHubService
from meltano.core.plugin import PluginType
from meltano.core.project import Project

from . import cli
from .params import pass_project
from .utils import InstrumentedCmd


@cli.command(
    cls=InstrumentedCmd,
    short_help="List the available plugins in Meltano Hub and their variants.",
)
@click.argument(
    "plugin_type", type=click.Choice([*list(PluginType), "all"]), default="all"
)
@pass_project()
@click.pass_context
def discover(ctx: click.Context, project: Project, plugin_type: str):
    """
    List the available discoverable plugins and their variants.

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface#discover
    """
    hub_service = MeltanoHubService(project)
    if plugin_type == "all":
        plugin_types = [
            plugin_type for plugin_type in list(PluginType) if plugin_type.discoverable
        ]
    else:
        plugin_types = [PluginType.from_cli_argument(plugin_type)]

    for idx, discovered_plugin_type in enumerate(plugin_types):
        if idx > 0:
            click.echo()

        click.secho(f"{str(discovered_plugin_type).capitalize()}", fg="green")

        try:
            plugin_type_index = hub_service.get_plugins_of_type(discovered_plugin_type)
        except Exception:
            click.secho(
                f"Can not retrieve {discovered_plugin_type} from the Hub",
                fg="yellow",
                err=True,
            )
            continue

        for plugin_name, plugin in plugin_type_index.items():
            click.echo(plugin_name, nl=False)

            if len(plugin.variants) > 1:
                click.echo(f", variants: {', '.join(plugin.variant_labels)}")
            else:
                click.echo()



"""Config management CLI."""

from __future__ import annotations

import asyncio
import json
import logging
import tempfile
from functools import wraps
from pathlib import Path
from typing import Any

import click
import dotenv

from meltano.cli import cli
from meltano.cli.interactive import InteractiveConfig
from meltano.cli.params import pass_project
from meltano.cli.utils import (
    CliEnvironmentBehavior,
    CliError,
    InstrumentedGroup,
    PartialInstrumentedCmd,
)
from meltano.core.db import project_engine
from meltano.core.plugin import PluginType
from meltano.core.plugin.error import PluginNotFoundError
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.plugin_test_service import PluginTestServiceFactory
from meltano.core.project import Project
from meltano.core.project_plugins_service import ProjectPluginsService
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.settings_service import SettingValueStore
from meltano.core.settings_store import StoreNotSupportedError
from meltano.core.tracking.contexts import CliEvent, PluginsTrackingContext

logger = logging.getLogger(__name__)


def _get_ctx_arg(*args: Any) -> click.core.Context:
    """Get the click.core.Context arg from a set of args.

    Args:
        args: the args to get Context from

    Returns:
        The click.core.Context arg.

    Raises:
        ValueError: if there is no click.core.Context in the given args.
    """
    for arg in args:
        if isinstance(arg, click.core.Context):
            return arg
    raise ValueError("No click.core.Context provided in *args")


def _get_store_choices() -> list[SettingValueStore]:
    """Get a list of valid choices for the --store flag.

    Returns:
        SettingValueStore.writables(), without meltano_env
    """
    writables = SettingValueStore.writables()
    writables.remove(SettingValueStore.MELTANO_ENV)
    return writables


def _use_meltano_env(func):
    """Override the 'meltano_yml' choice for a config command's 'store' argument.

    If an --environment flag is passed, the decorated command will use
    the MELTANO_ENV store instead of MELTANO_YML but _will not_ use MELTANO_ENV
    store if the active environment is set via the default environment.

    Args:
       func: the command to override

    Returns:
       A wrapped function with overridden store argument
    """

    @wraps(func)
    def _wrapper(*args, **kwargs):
        store = kwargs.pop("store")
        if store not in {SettingValueStore.MELTANO_YML, SettingValueStore.MELTANO_ENV}:
            return func(*args, **kwargs, store=store)
        ctx = _get_ctx_arg(*args)
        store = (
            SettingValueStore.MELTANO_YML
            if ctx.obj["is_default_environment"]
            else SettingValueStore.MELTANO_ENV
        )
        return func(*args, **kwargs, store=store)

    return _wrapper


def get_label(metadata) -> str:
    """Get the label for an environment variable's source.

    Args:
        metadata: the metadata for the variable

    Returns:
        string describing the source of the variable's value
    """
    source = metadata["source"]
    try:
        return f"from the {metadata['env_var']} variable in {source.label}"
    except KeyError:
        return f"from {source.label}"


@cli.group(
    cls=InstrumentedGroup,
    invoke_without_command=True,
    short_help="Display Meltano or plugin configuration.",
    environment_behavior=CliEnvironmentBehavior.environment_optional_ignore_default,
)
@click.option(
    "--plugin-type", type=click.Choice(PluginType.cli_arguments()), default=None
)
@click.argument("plugin_name")
@click.option(
    "--format",
    "config_format",
    type=click.Choice(["json", "env"]),
    default="json",
)
@click.option("--extras", is_flag=True, help="View or list only plugin extras.")
@pass_project(migrate=True)
@click.pass_context
def config(  # noqa: WPS231
    ctx,
    project: Project,
    plugin_type: str,
    plugin_name: str,
    config_format: str,
    extras: bool,
):
    """
    Display Meltano or plugin configuration.

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface#config
    """
    tracker = ctx.obj["tracker"]
    try:
        plugin_type = PluginType.from_cli_argument(plugin_type) if plugin_type else None
    except ValueError:
        tracker.track_command_event(CliEvent.aborted)
        raise

    plugins_service = ProjectPluginsService(project)

    try:
        plugin = plugins_service.find_plugin(
            plugin_name, plugin_type=plugin_type, configurable=True
        )
    except PluginNotFoundError:
        if plugin_name == "meltano":
            plugin = None
        else:
            tracker.track_command_event(CliEvent.aborted)
            raise

    if plugin:
        tracker.add_contexts(PluginsTrackingContext([(plugin, None)]))
    tracker.track_command_event(CliEvent.inflight)

    _, Session = project_engine(project)  # noqa: N806
    session = Session()
    try:
        if plugin:
            settings = PluginSettingsService(
                project,
                plugin,
                plugins_service=plugins_service,
            )
            invoker = PluginInvoker(project, plugin)
        else:
            settings = ProjectSettingsService(
                project, config_service=plugins_service.config_service
            )
            invoker = None

        ctx.obj["settings"] = settings
        ctx.obj["session"] = session
        ctx.obj["invoker"] = invoker

        if ctx.invoked_subcommand is None:
            if config_format == "json":
                process = extras is not True
                json_config = settings.as_dict(
                    extras=extras, process=process, session=session
                )
                click.echo(json.dumps(json_config, indent=2))
            elif config_format == "env":
                env = settings.as_env(extras=extras, session=session)

                with tempfile.NamedTemporaryFile() as temp_dotenv:
                    path = temp_dotenv.name
                    for key, value in env.items():
                        dotenv.set_key(path, key, value)

                    dotenv_content = Path(temp_dotenv.name).read_text()

                click.echo(dotenv_content)
    except Exception:
        tracker.track_command_event(CliEvent.failed)
        raise
    finally:
        session.close()


@_use_meltano_env
@config.command(
    cls=PartialInstrumentedCmd,
    name="list",
    short_help=(
        "List all settings for the specified plugin with their names, environment variables, and current values."
    ),
)
@click.option("--extras", is_flag=True)
@click.pass_context
def list_settings(ctx, extras: bool):
    """List all settings for the specified plugin with their names, environment variables, and current values."""
    settings = ctx.obj["settings"]
    session = ctx.obj["session"]
    tracker = ctx.obj["tracker"]

    printed_custom_heading = False
    printed_extra_heading = extras

    # If `--extras` is not specified (`False`), we still want to load both
    # regular and extra settings, since we show custom extras.
    load_extras = True if extras else None

    full_config = settings.config_with_metadata(session=session, extras=load_extras)
    for name, config_metadata in full_config.items():
        value = config_metadata["value"]
        source = config_metadata["source"]
        setting_def = config_metadata["setting"]

        if extras:
            if not setting_def.is_extra:
                continue

            if setting_def.is_custom and not printed_custom_heading:
                click.echo()
                click.echo("Custom:")
                printed_custom_heading = True
        elif setting_def.is_extra:
            if not setting_def.is_custom:
                continue

            if not printed_extra_heading:
                click.echo()
                click.echo("Custom extras, plugin-specific options handled by Meltano:")
                printed_extra_heading = True
        elif setting_def.is_custom and not printed_custom_heading:
            click.echo()
            click.echo("Custom, possibly unsupported by the plugin:")
            printed_custom_heading = True

        click.secho(name, fg="blue", nl=False)

        env_keys = [var.definition for var in settings.setting_env_vars(setting_def)]
        click.echo(f" [env: {', '.join(env_keys)}]", nl=False)

        if source is not SettingValueStore.DEFAULT:
            default_value = setting_def.value
            if default_value is not None:
                click.echo(f" (default: {default_value!r})", nl=False)

        if source is SettingValueStore.DEFAULT:
            label = "default"
        elif source is SettingValueStore.INHERITED:
            label = f"inherited from '{settings.plugin.parent.name}'"
        else:
            label = f"{get_label(config_metadata)}"

        current_value = click.style(f"{value!r}", fg="green")
        click.echo(f" current value: {current_value}", nl=False)

        unexpanded_value = config_metadata.get("unexpanded_value")
        if not unexpanded_value or unexpanded_value == value:
            click.echo(f" ({label})")
        else:
            click.echo(f" ({label}: {unexpanded_value!r})")

        if setting_def.description:
            click.echo("\t", nl=False)
            if setting_def.label:
                click.echo(f"{setting_def.label}: ", nl=False)
            click.echo(f"{setting_def.description}")

    docs_url = settings.docs_url
    if docs_url:
        click.echo()
        click.echo(
            f"To learn more about {settings.label} and its settings, visit {docs_url}"
        )
    tracker.track_command_event(CliEvent.completed)


@config.command(cls=PartialInstrumentedCmd)
@click.option(
    "--store",
    type=click.Choice(_get_store_choices()),
    default=SettingValueStore.AUTO,
)
@click.pass_context
@_use_meltano_env
def reset(ctx, store):
    """Clear the configuration (back to defaults)."""
    store = SettingValueStore(store)

    settings = ctx.obj["settings"]
    session = ctx.obj["session"]
    tracker = ctx.obj["tracker"]

    try:
        metadata = settings.reset(store=store, session=session)
    except StoreNotSupportedError as err:
        tracker.track_command_event(CliEvent.aborted)
        raise CliError(
            f"{settings.label.capitalize()} settings in {store.label} could not be reset: {err}"
        ) from err

    store = metadata["store"]
    click.secho(
        f"{settings.label.capitalize()} settings in {store.label} were reset",
        fg="green",
    )
    tracker.track_command_event(CliEvent.completed)


@config.command(cls=PartialInstrumentedCmd, name="set")
@click.option("--interactive", is_flag=True)
@click.argument("setting_name", nargs=-1)
@click.argument("value", required=False)
@click.option(
    "--store",
    type=click.Choice(_get_store_choices()),
    default=SettingValueStore.AUTO,
)
@click.pass_context
@_use_meltano_env
def set_(ctx, setting_name, value, store, interactive):
    """Set the configurations' setting `<name>` to `<value>`."""
    interaction = InteractiveConfig(ctx=ctx, store=store, extras=False)
    if interactive:
        interaction.configure_all()
    else:
        interaction.set_value(setting_name=setting_name, value=value, store=store)


@config.command(cls=PartialInstrumentedCmd, name="test")
@click.pass_context
def test(ctx):
    """Test the configuration of a plugin."""
    invoker = ctx.obj["invoker"]
    tracker = ctx.obj["tracker"]
    if not invoker:
        tracker.track_command_event(CliEvent.aborted)
        raise CliError("Testing of the Meltano project configuration is not supported")

    session = ctx.obj["session"]

    async def _validate():  # noqa: WPS430
        plugin_test_service = PluginTestServiceFactory(invoker).get_test_service()
        async with plugin_test_service.plugin_invoker.prepared(session):
            return await plugin_test_service.validate()

    try:
        is_valid, detail = asyncio.run(_validate())
    except Exception:
        tracker.track_command_event(CliEvent.failed)
        raise

    if not is_valid:
        tracker.track_command_event(CliEvent.failed)
        raise CliError("\n".join(("Plugin configuration is invalid", detail)))

    click.secho("Plugin configuration is valid", fg="green")
    tracker.track_command_event(CliEvent.completed)


@config.command(cls=PartialInstrumentedCmd)
@click.argument("setting_name", nargs=-1, required=True)
@click.option(
    "--store",
    type=click.Choice(_get_store_choices()),
    default=SettingValueStore.AUTO,
)
@click.pass_context
@_use_meltano_env
def unset(ctx, setting_name, store):
    """Unset the configurations' setting called `<name>`."""
    store = SettingValueStore(store)

    settings = ctx.obj["settings"]
    session = ctx.obj["session"]
    tracker = ctx.obj["tracker"]

    path = list(setting_name)
    try:
        metadata = settings.unset(path, store=store, session=session)
    except StoreNotSupportedError as err:
        tracker.track_command_event(CliEvent.aborted)
        raise CliError(
            f"{settings.label.capitalize()} setting '{path}' in {store.label} could not be unset: {err}"
        ) from err

    name = metadata["name"]
    store = metadata["store"]
    click.secho(
        f"{settings.label.capitalize()} setting '{name}' in {store.label} was unset",
        fg="green",
    )

    current_value, source = settings.get_with_source(name, session=session)
    if source is not SettingValueStore.DEFAULT:
        click.secho(
            f"Current value is now: {current_value!r} ({get_label(metadata)})",
            fg="yellow",
        )
    tracker.track_command_event(CliEvent.completed)



"""StateStoreManager for Azure Blob storage backend."""
from __future__ import annotations

import re

from meltano.core.state_store.filesystem import BaseFilesystemStateStoreManager


class AZStorageStateStoreManager(BaseFilesystemStateStoreManager):
    """State backend for Azure Blob Storage."""

    label: str = "Azure Blob Storage"

    def __init__(
        self,
        container_name: str | None = None,
        connection_string: str | None = None,
        prefix: str | None = None,
        **kwargs,
    ):
        """Initialize the BaseFilesystemStateStoreManager.

        Args:
            container_name: the container to store state in.
            connection_string: connection string to use in authenticating to Azure
            prefix: the prefix to store state at
            kwargs: additional keyword args to pass to parent
        """
        super().__init__(**kwargs)
        self.connection_string = connection_string
        self.container_name = container_name or self.parsed.hostname
        self.prefix = prefix or self.parsed.path
        self._client = None

    @staticmethod
    def is_file_not_found_error(err: Exception) -> bool:
        """Check if err is equivalent to file not being found.

        Args:
            err: the err to check

        Returns:
            True if error represents file not being found, else False
        """
        from azure.core.exceptions import ResourceNotFoundError  # type: ignore

        return (
            isinstance(err, ResourceNotFoundError)
            and "ErrorCode:BlobNotFound" in err.args[0]
        )

    @property
    def client(self):
        """Get an authenticated azure.storage.blob.BlobServiceClient.

        Returns:
            An authenticated azure.storage.blob.BlobServiceClient
        """
        if self.connection_string and not self._client:
            from azure.storage.blob import BlobServiceClient  # type: ignore

            self._client = BlobServiceClient.from_connection_string(
                self.connection_string
            )
        return self._client

    @property
    def state_dir(self) -> str:
        """Get the prefix that state should be stored at.

        Returns:
            The relevant prefix
        """
        return self.prefix.lstrip(self.delimiter).rstrip(self.delimiter)

    def get_state_ids(self, pattern: str | None = None):  # noqa: WPS210
        """Get list of state_ids stored in the backend.

        Args:
            pattern: glob-style pattern to filter state_ids by

        Returns:
            List of state_ids
        """
        if pattern:
            pattern_re = re.compile(pattern.replace("*", ".*"))
        state_ids = set()
        container_client = self.client.get_container_client(self.container_name)
        for blob in container_client.list_blobs(
            name_starts_with=self.prefix.lstrip("/")
        ):
            (state_id, filename) = blob.name.split("/")[-2:]
            if filename == "state.json":
                if not pattern:
                    state_ids.add(state_id)
                elif pattern_re.match(state_id):
                    state_ids.add(state_id)
        return list(state_ids)

    def delete(self, file_path: str):
        """Delete the file/blob at the given path.

        Args:
            file_path: the path to delete.

        Raises:
            Exception: if error not indicating file is not found is thrown
        """
        blob_client = self.client.get_blob_client(
            container=self.container_name, blob=file_path
        )
        try:
            blob_client.delete_blob()
        except Exception as e:
            if self.is_file_not_found_error(e):
                ...
            else:
                raise e



from __future__ import annotations

from sqlalchemy import Column, types

from .models import SystemModel


class Setting(SystemModel):
    __tablename__ = "plugin_settings"

    # represent the mapping to the ENV
    label = Column(types.String)
    description = Column(types.Text)

    # represent a materialized path to support
    # a nested configuration.
    name = Column(types.String, primary_key=True)
    namespace = Column(types.String, primary_key=True, nullable=True)
    value = Column(types.PickleType)
    enabled = Column(types.Boolean, default=False)

    def __repr__(self):
        enabled_marker = "E" if self.enabled else ""
        return f"<({self.namespace}) {self.name}={self.value} {enabled_marker}>"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['meltano',
 'meltano.api',
 'meltano.api.controllers',
 'meltano.api.events',
 'meltano.api.executor',
 'meltano.api.models',
 'meltano.api.security',
 'meltano.api.workers',
 'meltano.cli',
 'meltano.cli.interactive',
 'meltano.core',
 'meltano.core.behavior',
 'meltano.core.block',
 'meltano.core.bundle',
 'meltano.core.container',
 'meltano.core.hub',
 'meltano.core.job',
 'meltano.core.logging',
 'meltano.core.plugin',
 'meltano.core.plugin.dbt',
 'meltano.core.plugin.singer',
 'meltano.core.runner',
 'meltano.core.state_store',
 'meltano.core.tracking',
 'meltano.core.tracking.contexts',
 'meltano.core.utils',
 'meltano.migrations',
 'meltano.migrations.utils',
 'meltano.migrations.versions',
 'meltano.oauth']

package_data = \
{'': ['*'],
 'meltano.api': ['static/*',
                 'static/css/*',
                 'static/js/*',
                 'static/logos/*',
                 'templates/*',
                 'templates/email/pipeline_manual_run/*',
                 'templates/security/*',
                 'templates/security/email/*'],
 'meltano.core.tracking': ['iglu-client-embedded/schemas/com.meltano/block_event/jsonschema/*',
                           'iglu-client-embedded/schemas/com.meltano/cli_context/jsonschema/*',
                           'iglu-client-embedded/schemas/com.meltano/cli_event/jsonschema/*',
                           'iglu-client-embedded/schemas/com.meltano/environment_context/jsonschema/*',
                           'iglu-client-embedded/schemas/com.meltano/exception_context/jsonschema/*',
                           'iglu-client-embedded/schemas/com.meltano/exit_event/jsonschema/*',
                           'iglu-client-embedded/schemas/com.meltano/plugins_context/jsonschema/*',
                           'iglu-client-embedded/schemas/com.meltano/project_context/jsonschema/*',
                           'iglu-client-embedded/schemas/com.meltano/telemetry_state_change_event/jsonschema/*'],
 'meltano.oauth': ['templates/*']}

install_requires = \
['aiodocker>=0.21.0,<0.22.0',
 'aiohttp>=3.4.4,<4.0.0',
 'alembic>=1.5.0,<2.0.0',
 'atomicwrites>=1.2.1,<2.0.0',
 'authlib>=1.0.1,<2.0.0',
 'backoff>=2.1.2,<3.0.0',
 'bcrypt>=3.2.0,<4.0.0',
 'cached-property>=1,<2',
 'click-default-group>=1.2.1,<2.0.0',
 'click>=8.1,<9.0',
 'croniter>=1.3.5,<2.0.0',
 'email-validator>=1.1.2,<2.0.0',
 'fasteners>=0.17.3,<0.18.0',
 'flask-cors>=3.0.7,<4.0.0',
 'flask-executor>=0.10,<0.11',
 'flask-login==0.6.1',
 'flask-restful>=0.3.7,<0.4.0',
 'flask-sqlalchemy>=2.4.4,<3.0.0',
 'flask>=2.1,<3.0',
 'flatten-dict>=0,<1',
 'gunicorn>=20.1.0,<21.0.0',
 'jsonschema>=4.9,<5.0',
 'meltano-flask-security>=0.1.0,<0.2.0',
 'packaging>=21.3,<22.0',
 'psutil>=5.6.3,<6.0.0',
 'psycopg2-binary>=2.8.5,<3.0.0',
 'pyhumps>=3.0.0,<4.0.0',
 'python-dotenv>=0.20.0,<0.21.0',
 'python-gitlab>=3.5.0,<4.0.0',
 'pyyaml>=6.0.0,<7.0.0',
 'requests>=2.23.0,<3.0.0',
 'rich>=12.5.1,<13.0.0',
 'ruamel.yaml>=0.17.21,<0.18.0',
 'smart-open>=6.2.0,<7.0.0',
 'smtpapi>=0.4.1,<0.5.0',
 'snowplow-tracker>=0.10.0,<0.11.0',
 'sqlalchemy>=1.3.19,<2.0.0',
 'structlog>=21.2.0,<22.0.0',
 'tzlocal>=4.2.0,<5.0.0',
 'uvicorn[standard]>=0.17.6,<0.18.0',
 'werkzeug>=2.1,<=2.1.3']

extras_require = \
{'azure': ['azure-common>=1.1.28,<2.0.0',
           'azure-core>=1.26.0,<2.0.0',
           'azure-storage-blob>=12.14.1,<13.0.0'],
 'gcs': ['google-cloud-storage>=1.31.0'],
 'mssql': ['pymssql>=2.2.5,<3.0.0'],
 's3': ['boto3>=1.25.3,<2.0.0']}

entry_points = \
{'console_scripts': ['meltano = meltano.cli:main']}

setup_kwargs = {
    'name': 'meltano',
    'version': '2.12.0',
    'description': 'Meltano: Your DataOps Platform Infrastructure',
    'long_description': '\n<h1 align="center">Meltano - Your CLI for ELT+</h1>\n<h3 align="center">Open source, flexible, scales to your needs. Confidently move, transform, and test your data using tools you know with a data engineering workflow you’ll love.</h3>\n\n<div align="center">\n<a href="https://github.com/codespaces/new?template_repository=meltano/meltano-codespace-ready">\n<img alt="Try codespaces" src="https://img.shields.io/static/v1?label=&message=Try live demo with Codespaces&color=02a5a5&style=for-the-badge&logo=github"/>\n</a>\n</div>\n\n---\n\n![Meltano Logo](https://lh4.googleusercontent.com/WHoN-WpacMaVicq-jRuIvCQjCIdPZwYOwBgd38k9JjMpX1Z7THUqowY-oRsTzGUbAvb8F4tcb9BJYyfX9MeA2ECirsWZ7XBHteDZ_y59REMwHjq1AX05U2k8H6mdI4G_olF27gadCfp1Wx7cVQ)\n\n<div align="center">\n<a href="https://docs.meltano.com/">\n<img alt="Docs" src="https://img.shields.io/website?down_color=red&down_message=offline&label=Docs&up_color=blue&up_message=online&url=https%3A%2F%2Fdocs.meltano.com%2F"/>\n</a>\n<a href="https://github.com/meltano/meltano/actions/workflows/test.yml?query=branch%3Amain">\n<img alt="Tests" src="https://github.com/meltano/meltano/actions/workflows/test.yml/badge.svg"/>\n</a>\n<a href="https://codecov.io/github/meltano/meltano">\n<img alt="Codecov" src="https://codecov.io/gh/meltano/meltano/branch/main/graph/badge.svg"/>\n</a>\n<a href="https://libraries.io/pypi/meltano/sourcerank">\n<img alt="Libraries.io SourceRank" src="https://img.shields.io/librariesio/sourcerank/pypi/meltano?label=SourceRank"/>\n</a>\n<a href="https://libraries.io/pypi/meltano">\n<img alt="Libraries.io dependency status for latest release" src="https://img.shields.io/librariesio/release/pypi/meltano?label=Dependencies"/>\n</a>\n</div>\n\n<div align="center">\n<a href="https://github.com/meltano/meltano/graphs/contributors">\n<img alt="GitHub contributors" src="https://img.shields.io/github/contributors/meltano/meltano?label=Contributors"/>\n</a>\n<a href="https://github.com/meltano/meltano/blob/main/LICENSE">\n<img alt="GitHub" src="https://img.shields.io/github/license/meltano/meltano?color=blue&label=License"/>\n</a>\n<a href="https://pypi.org/project/meltano/">\n<img alt="Meltano Python Package Version" src="https://img.shields.io/pypi/v/meltano?label=Version"/>\n</a>\n<a href="https://pypi.org/project/meltano/">\n<img alt="Supported Python Versions" src="https://img.shields.io/pypi/pyversions/meltano?label=Python"/>\n</a>\n<a href="https://pypi.org/project/meltano/">\n<img alt="Monthly PyPI Downloads" src="https://img.shields.io/pypi/dm/meltano?label=PyPI%20Downloads"/>\n</a>\n<a href="https://hub.docker.com/r/meltano/meltano">\n<img alt="Docker Pulls" src="https://img.shields.io/docker/pulls/meltano/meltano?label=Docker%20Pulls"/>\n</a>\n</div>\n\n---\n\nWelcome to your CLI for ELT+. It\'s open source, flexible, scales to your needs. Confidently move, transform, and test your data using tools you know with a data engineering workflow you’ll love.\n\nIf you\'re a fan, star the repo ⭐️. [Plus this month, every ⭐ on GitHub removes 2 lb/~1 kg of trash from our waterways 🌊](https://meltano.com/blog/extracting-trash-to-transform-our-waterways/)\n\n\n\nIntegrations\n------------\n\n[MeltanoHub](https://hub.meltano.com/) is the single source of truth to find any Meltano plugins as well as [Singer](https://singer.io/) taps and targets. Users are also able to add more plugins to the Hub and have them immediately discoverable and usable within Meltano. The Hub is lovingly curated by Meltano and the wider Meltano community.\n\nInstallation\n------------\n\nIf you\'re ready to build your ideal data platform and start running data workflows across multiple tools, start by following the [Installation guide](https://docs.meltano.com/getting-started/installation) to have Meltano up and running in your device.\n\nDocumentation\n-------------\n\nCheck out the ["Getting Started" guide](https://docs.meltano.com/getting-started) or find the full documentation at [https://docs.meltano.com](https://docs.meltano.com/).\n\nContributing\n------------\n\nMeltano is a truly open-source project, built for and by its community. We happily welcome and encourage your contributions. Start by browsing through our [issue tracker](https://github.com/meltano/meltano/issues?q=is%3Aopen+is%3Aissue) to add your ideas to the roadmap. If you\'re still unsure on what to contribute at the moment, you can always check out the list of open issues labeled as "[Accepting Merge Requests](https://github.com/meltano/meltano/issues?q=is%3Aopen+is%3Aissue+label%3A%22accepting+merge+requests%22)".\n\nFor more information on how to contribute to Meltano, refer to our [contribution guidelines](https://docs.meltano.com/contribute/).\n\nCommunity\n---------\n\nWe host weekly online events where you can engage with us directly. Check out more information in our [Community](https://meltano.com/community/) page.\n\nIf you have any questions, want sneak peeks of features or would just like to say hello and network, join our community of over +2,500 data professionals!\n\n👋 [Join us on Slack!](https://meltano.com/slack)\n\nResponsible Disclosure Policy\n-----------------------------\n\nPlease refer to the [responsible disclosure policy](https://docs.meltano.com/the-project/responsible-disclosure) on our website.\n\nLicense\n-------\n\nThis code is distributed under the MIT license, see the [LICENSE](https://github.com/meltano/meltano/blob/main/LICENSE) file.\n',
    'author': 'Meltano',
    'author_email': 'hello@meltano.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://meltano.com',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



"""Interactive configuration handler."""

from __future__ import annotations

import click
from jinja2 import BaseLoader, Environment
from rich.console import Console, Group
from rich.markdown import Markdown
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from meltano.cli.interactive.utils import InteractionStatus
from meltano.cli.utils import CliError
from meltano.core.environment_service import EnvironmentService
from meltano.core.project import Project
from meltano.core.settings_service import (
    REDACTED_VALUE,
    SettingKind,
    SettingsService,
    SettingValueStore,
)
from meltano.core.settings_store import StoreNotSupportedError
from meltano.core.tracking.contexts import CliEvent

PLUGIN_COLOR = "magenta"
ENVIRONMENT_COLOR = "orange1"
SETTING_COLOR = "blue1"
VALUE_COLOR = "green"

HOME_SCREEN_TEMPLATE = """[bold underline]Configuring [{{ plugin_color }}]{{ plugin_name.capitalize() | safe }}[/{{ plugin_color }}] {% if environment_name %}in Environment[{{ environment_color }}]{{ environment_name }}[/{{ environment_color }}] {% endif %}Interactively[/bold underline]

Following the prompts below, you will be guided through configuration of this plugin.

Meltano is responsible for managing the configuration of all of a project’s plugins.
It knows what settings are supported by each plugin, and how and when different types of plugins expect to be fed that configuration.

To determine the values of settings, Meltano will look in 4 main places, with each taking precedence over the next:

  1. Environment variables
  2. Your meltano.yml project file
  3. Your project's system database
  4. The default values set in the plugin's settings metadata

Within meltano.yml you can also associate configuration with a Meltano Environment, allowing you to define custom layers of configuration within your project.

To learn more about configuration options, see the [link=https://docs.meltano.com/guide/configuration]Meltano Configuration Guide[/link]

[bold underline]Settings[/bold underline]
{% for setting in settings %}
{{ loop.index }}. [blue]{{ setting["name"] }}[/blue]: {{ setting["description"] | safe }}
{%- endfor %}

{% if plugin_url %}To learn more about {{ plugin_name | safe }} and its settings, visit [link={{ plugin_url }}]{{ plugin_url }}[/link]{% endif %}
"""


class InteractiveConfig:  # noqa: WPS230, WPS214
    """Manage Config interactively."""

    def __init__(self, ctx, store, extras=False, max_width=None):
        """Initialise InteractiveConfig instance."""
        self.ctx = ctx
        self.store = store
        self.extras = extras
        self.project: Project = self.ctx.obj["project"]
        self.settings: SettingsService = self.ctx.obj["settings"]
        self.session = self.ctx.obj["session"]
        self.tracker = self.ctx.obj["tracker"]
        self.environment_service = EnvironmentService(self.project)
        self.max_width = max_width or 75  # noqa: WPS432
        self.console = Console()

    @property
    def configurable_settings(self):
        """Return settings available for interactive configuration."""
        return self.settings.config_with_metadata(
            session=self.session, extras=self.extras, redacted=True
        )

    @property
    def setting_choices(self):
        """Return simplified setting choices, for easy printing."""
        setting_choices = []
        for index, (name, config_metadata) in enumerate(
            self.configurable_settings.items()
        ):
            description = config_metadata["setting"].description
            description = "" if description is None else description
            setting_choices.append((str(index + 1), name, description))
        return setting_choices

    def truncate(self, text: str) -> str:
        """Truncate text."""
        if len(text) >= self.max_width:
            return f"{text[: self.max_width - 3]}..."
        return text

    def _print_home_screen(self):
        """Print screen for this interactive."""
        markdown_template = Environment(loader=BaseLoader, autoescape=True).from_string(
            HOME_SCREEN_TEMPLATE
        )
        markdown_text = markdown_template.render(
            {
                "plugin_color": PLUGIN_COLOR,
                "environment_color": ENVIRONMENT_COLOR,
                "setting_color": SETTING_COLOR,
                "plugin_name": self.settings.label,
                "plugin_url": self.settings.docs_url,
                "environment_name": self.project.active_environment.name
                if self.project.active_environment
                else None,
                "settings": [
                    {
                        "name": name,
                        "description": self.truncate(description.replace("\n", " ")),
                    }
                    for _, name, description in self.setting_choices
                ],
            }
        )
        self.console.print(Panel(Text.from_markup(markdown_text)))

    def _print_setting(self, name, config_metadata, index, last_index):
        """Print setting."""
        value = config_metadata["value"]
        source = config_metadata["source"]
        setting_def = config_metadata["setting"]
        details = Table(show_header=False)
        details.add_column("name", justify="right")
        details.add_column("value")

        pre = [
            Text.from_markup(
                f"[bold underline][{PLUGIN_COLOR}]{self.settings.label.capitalize()}[/{PLUGIN_COLOR}][/bold underline] Setting {index} of {last_index}"
            )
        ]

        if setting_def.is_extra:
            pre.append(
                Text.from_markup(
                    "[yellow1]Custom Extra: plugin-specific options handled by Meltano[/yellow1]"
                )
            )

        elif setting_def.is_custom:
            pre.append(
                Text.from_markup(
                    "[yellow1]Custom Setting: possibly unsupported by the plugin[/yellow1]"
                )
            )

        details.add_row(
            Text("Name"), Text.from_markup(f"[{SETTING_COLOR}]{name}[/{SETTING_COLOR}]")
        )

        if source is SettingValueStore.DEFAULT:
            label = "default"
        elif source is SettingValueStore.INHERITED:
            label = f"inherited from '{self.settings.plugin.parent.name}'"
        else:
            label = f"from {source.label}"
        expanded_value = value if value is not None else "(empty string)"
        unexpanded_value = config_metadata.get("unexpanded_value")
        if unexpanded_value:
            current_value = (
                unexpanded_value if unexpanded_value is not None else "(empty string)"
            )

            details.add_row(Text("Current Expanded Value"), Text(f"{expanded_value}"))
        else:
            current_value = value if value is not None else "(empty string)"
        details.add_row(
            Text(f"Current Value ({label})"),
            Text.from_markup(f"[{VALUE_COLOR}]{current_value}[/{VALUE_COLOR}]"),
        )

        if setting_def.kind:
            details.add_row(Text("Kind"), Text(f"{setting_def.kind}"))
        if source is not SettingValueStore.DEFAULT:
            default_value = setting_def.value
            if default_value is not None:
                details.add_row(Text("Default"), Text(f"{default_value!r}"))
        env_keys = [
            var.definition for var in self.settings.setting_env_vars(setting_def)
        ]

        details.add_row(Text("Env(s)"), Text(f"{', '.join(env_keys)}"))
        post = []
        if setting_def.description:
            post.append(
                Group(
                    Text(" Description:"),
                    Panel(Markdown(setting_def.description, justify="left")),
                )
            )

        docs_url = self.settings.docs_url
        if docs_url:
            post.append(
                Text.from_markup(
                    f" To learn more about {self.settings.label} and its settings, visit [link={docs_url}]{docs_url}[/link]"
                )
            )

        self.console.print(Panel(Group(*pre, details, *post)))

    @staticmethod
    def _value_prompt(config_metadata):
        if config_metadata["setting"].kind != SettingKind.OPTIONS:
            return (
                click.prompt(
                    "New value",
                    default="",
                    show_default=False,
                    hide_input=True,
                    confirmation_prompt=True,
                )
                if config_metadata["setting"].is_redacted
                else click.prompt("New value", default="", show_default=False)
            )

        options_index = {
            str(index + 1): value
            for index, value in enumerate(
                (chs["label"], chs["value"])
                for chs in config_metadata["setting"].options
            )
        }

        click.echo()
        for index, value in options_index.items():
            click.echo(f"{index}. {value[0]}")
        click.echo()
        chosen_index = click.prompt(
            "Select value",
            type=click.Choice(list(options_index.keys())),
            show_default=False,
        )
        return options_index[chosen_index][1]

    def configure(self, name, index=None, last_index=None, show_set_prompt=True):
        """Configure a single setting interactively."""
        config_metadata = next(
            (
                config_metadata
                for nme, config_metadata in self.configurable_settings.items()
                if nme == name
            )
        )
        self._print_setting(
            name=name,
            config_metadata=config_metadata,
            index=index,
            last_index=last_index,
        )

        action = "y"
        if show_set_prompt:
            try:
                click.echo()
                action = click.prompt(
                    "Set this value (Y/n) or exit (e)?",
                    default="y",
                    type=click.Choice(["y", "n", "e"], case_sensitive=False),
                )
            except click.Abort:
                action = "e"

        if action.lower() == "y":
            while True:
                click.echo()
                try:
                    new_value = self._value_prompt(config_metadata)
                except click.Abort:
                    click.echo()
                    click.echo("Skipping...")
                    click.pause()
                    return InteractionStatus.SKIP

                try:
                    click.echo()
                    self.set_value(
                        setting_name=tuple(name.split(".")),
                        value=new_value,
                        store=self.store,
                        interactive=True,
                    )
                    click.echo()
                    click.pause()
                    return InteractionStatus.SKIP
                except Exception as e:
                    self.tracker.track_command_event(CliEvent.inflight)
                    click.secho(f"Failed to set value: {e}", fg="red")

        elif action.lower() == "n":
            return InteractionStatus.SKIP

        elif action.lower() == "e":
            return InteractionStatus.EXIT

    def configure_all(self):
        """Configure all settings."""
        numeric_choices = [idx for idx, _, _ in self.setting_choices]
        if not numeric_choices:
            click.secho(
                "There are no settings to configure. "
                "For help, please see https://melta.no#no-plugin-settings-defined",
                fg="yellow",
            )
            self.tracker.track_command_event(CliEvent.completed)
            return

        while True:
            click.clear()
            self._print_home_screen()
            choices = ["all", *numeric_choices, "e"]

            branch = "all"
            try:
                click.echo()
                branch = click.prompt(
                    "Loop through all settings (all), select a setting by "
                    f"number ({min(int(chs) for chs in numeric_choices)} - "
                    f"{max(int(chs) for chs in numeric_choices)}), or exit (e)?",
                    type=click.Choice(choices, case_sensitive=False),
                    default="all",
                    show_choices=False,
                )
            except click.Abort:
                click.echo()
                branch = "e"

            if branch == "all":
                for index, name, _ in self.setting_choices:
                    click.clear()
                    status = InteractionStatus.START
                    while status not in {
                        InteractionStatus.SKIP,
                        InteractionStatus.EXIT,
                    }:
                        status = self.configure(
                            name=name,
                            index=index,
                            last_index=len(self.setting_choices),
                        )
                    if status == InteractionStatus.EXIT:
                        break
            elif branch.lower() == "e":
                self.tracker.track_command_event(CliEvent.completed)
                click.echo()
                return
            else:
                choice_name = next(
                    nme for idx, nme, _ in self.setting_choices if idx == branch
                )
                click.clear()
                status = self.configure(
                    name=choice_name,
                    index=branch,
                    last_index=len(self.setting_choices),
                    show_set_prompt=False,
                )

    def set_value(self, setting_name, value, store, interactive=False):
        """Set value helper function."""
        settings = self.settings
        path = list(setting_name)
        try:
            value, metadata = settings.set_with_metadata(
                path, value, store=store, session=self.session
            )
        except StoreNotSupportedError as err:
            if interactive:
                self.tracker.track_command_event(CliEvent.inflight)
            else:
                self.tracker.track_command_event(CliEvent.aborted)
            raise CliError(
                f"{settings.label.capitalize()} setting '{path}' could not be set in {store.label}: {err}"
            ) from err

        name = metadata["name"]
        store = metadata["store"]
        is_redacted = metadata["setting"] and metadata["setting"].is_redacted
        if is_redacted:
            value = REDACTED_VALUE
        click.secho(
            f"{settings.label.capitalize()} setting '{name}' was set in {store.label}: {value!r}",
            fg=VALUE_COLOR,
        )

        current_value, source = settings.get_with_source(name, session=self.session)
        if source != store:
            if is_redacted:
                current_value = REDACTED_VALUE
            click.secho(
                f"Current value is still: {current_value!r} (from {source.label})",
                fg="yellow",
            )

        if interactive:
            self.tracker.track_command_event(CliEvent.inflight)
        else:
            self.tracker.track_command_event(CliEvent.completed)



"""Defines MeltanoInvoker."""

from __future__ import annotations

import os
import platform
import subprocess
import sys
from collections.abc import Mapping
from pathlib import Path

from meltano.core.project import Project
from meltano.core.project_settings_service import (
    ProjectSettingsService,
    SettingValueStore,
)
from meltano.core.tracking import Tracker

MELTANO_COMMAND = "meltano"


class MeltanoInvoker:
    """Class used to find and invoke all commands passed to it."""

    def __init__(
        self, project: Project, settings_service: ProjectSettingsService | None = None
    ):
        """
        Load the class with the project and service settings.

        Args:
            project: Project class
            settings_service: ProjectSettingsService Class blank
        """
        self.project = project
        self.tracker = Tracker(project)
        self.settings_service = settings_service or ProjectSettingsService(project)

    def invoke(self, args, command=MELTANO_COMMAND, env=None, **kwargs):
        """
        Invoke meltano or other provided command.

        Args:
            args: list.
            command: string containing command to invoke.
            env: dictionary.
            kwargs: dictionary.

        Returns:
            A CompletedProcess class object from subprocess.run().
        """
        return subprocess.run(
            [self._executable_path(command), *args],
            **kwargs,
            env=self._executable_env(env),
        )

    def _executable_path(self, command):
        if command == MELTANO_COMMAND:
            # This symlink is created by Project.activate
            executable_symlink = self.project.run_dir().joinpath("bin")
            if executable_symlink.exists():
                return str(executable_symlink)

        executable = Path(
            os.path.dirname(sys.executable),
            f"{command}.exe" if platform.system() == "Windows" else command,
        )

        # Fall back on expecting command to be in the PATH
        return str(executable) if executable.exists() else command

    def _executable_env(self, env: Mapping[str, str] | None = None) -> dict[str, str]:
        if env is None:
            env = {}
        return {
            # Include env that project settings are evaluated in
            **self.settings_service.env,
            # Include env for settings explicitly overridden using CLI flags
            **self.settings_service.as_env(source=SettingValueStore.CONFIG_OVERRIDE),
            # Include explicitly provided env
            **env,
            # Include telemetry env vars
            **self.tracker.env,
        }



"""API Orchestation of Pipellines and Validations."""

from __future__ import annotations

import asyncio
import logging
import shutil

import sqlalchemy
from flask import Response, jsonify, request, send_file
from flask_restful import Api, Resource, fields, marshal_with
from werkzeug.exceptions import Conflict, UnprocessableEntity

from meltano.api.api_blueprint import APIBlueprint
from meltano.api.executor import run_schedule
from meltano.api.json import freeze_keys
from meltano.api.models import db
from meltano.api.models.subscription import Subscription
from meltano.api.security.auth import block_if_readonly
from meltano.core.behavior.canonical import Canonical
from meltano.core.job import JobFinder
from meltano.core.logging import (
    JobLoggingService,
    MissingJobLogException,
    SizeThresholdJobLogException,
)
from meltano.core.plugin import PluginRef
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.plugin_discovery_service import PluginNotFoundError
from meltano.core.plugin_invoker import invoker_factory
from meltano.core.plugin_test_service import PluginTestServiceFactory
from meltano.core.project import Project
from meltano.core.project_plugins_service import ProjectPluginsService
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.schedule_service import (
    ScheduleAlreadyExistsError,
    ScheduleDoesNotExistError,
    ScheduleService,
)
from meltano.core.setting_definition import SettingKind
from meltano.core.settings_service import FeatureFlags
from meltano.core.utils import slugify

from .errors import InvalidFileNameError
from .upload_helper import InvalidFileSizeError, InvalidFileTypeError, UploadHelper
from .utils import enforce_secure_filename


def get_config_with_metadata(settings):
    """Get configuration including metadata.

    Args:
        settings: List of settings to obtain metadata for.

    Returns:
        Configuration with setting metadata
    """
    config_dict = {}
    config_metadata = {}

    config_with_metadata = settings.config_with_metadata(
        extras=False, redacted=True, session=db.session
    )
    for key, metadata in config_with_metadata.items():
        config_dict[key] = metadata.pop("value")

        metadata.pop("setting", None)

        try:
            metadata["source_label"] = metadata["source"].label
            metadata["auto_store_label"] = metadata["auto_store"].label
        except KeyError:
            pass

        config_metadata[key] = metadata

    return {
        "config": freeze_keys(config_dict),
        "config_metadata": freeze_keys(config_metadata),
    }


def validate_plugin_config(
    plugin: PluginRef, name, value, project: Project, settings: PluginSettingsService
):
    """Check the plugin configuration.

    Args:
        plugin: PluginRef class.
        name: Name of the plugin to validate
        value: Name of plugin setting file
        project: Project class.
        settings: PluginSettingsService class.

    Returns:
        A boolean true for pass, false for fail.
    """
    setting_def = settings.find_setting(name)
    # we want to prevent the edition of protected settings from the UI
    if setting_def.protected:
        logging.warning("Cannot set a 'protected' setting externally.")
        return False

    if setting_def.kind == SettingKind.FILE and value and value != "":
        uploads_directory = project.extract_dir(plugin.name)
        resolved_file_path = project.root_dir(value).resolve()
        if not str(resolved_file_path).startswith(f"{uploads_directory}/"):
            logging.warning(
                "Cannot set a file configuration to a path outside the project directory"
            )
            return False

    old_value, metadata = settings.get_with_metadata(name, session=db.session)
    if not metadata["overwritable"]:
        logging.warning("Cannot overwrite this setting.")
        return False

    return True


orchestrations_bp = APIBlueprint("orchestrations", __name__)  # noqa: N816
orchestrationsAPI = Api(  # noqa: N816
    orchestrations_bp,
    errors={
        "UnprocessableEntity": {
            "error": True,
            "code": "The subscription could not be created.",
            "status": UnprocessableEntity.code,
        },
        "Conflict": {
            "error": True,
            "code": "A subscription already exists for this address.",
            "status": Conflict.code,
        },
    },
)


@orchestrations_bp.errorhandler(ScheduleAlreadyExistsError)  # noqa: F811
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"A pipeline with the name '{ex.schedule.name}' already exists. Try renaming the pipeline.",
            }
        ),
        409,
    )


@orchestrations_bp.errorhandler(ScheduleDoesNotExistError)  # noqa: F811
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"A pipeline with the name '{ex.name}' does not exist..",
            }
        ),
        404,
    )


@orchestrations_bp.errorhandler(InvalidFileNameError)  # noqa: F811
def _handle(ex):
    return (jsonify({"error": True, "code": "The file lacks a valid name."}), 400)


@orchestrations_bp.errorhandler(InvalidFileTypeError)  # noqa: F811
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"The file '{ex.file.filename}' must be one of the following types: {ex.extensions}",
            }
        ),
        400,
    )


@orchestrations_bp.errorhandler(InvalidFileSizeError)  # noqa: F811
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"The file '{ex.file.filename}' is empty or exceeds the {ex.max_file_size} size limit.",
            }
        ),
        400,
    )


@orchestrations_bp.errorhandler(MissingJobLogException)  # noqa: F811
def _handle(ex):
    return (jsonify({"error": False, "code": str(ex)}), 204)


@orchestrations_bp.route("/jobs/state", methods=["POST"])
def job_state() -> Response:
    """Endpoint for getting the status of N jobs.

    Returns:
        A JSON containing all the endpoint job status for the N jobs.
    """
    poll_payload = request.get_json()
    state_ids = poll_payload["state_ids"]

    jobs = []
    for state_id in state_ids:
        finder = JobFinder(state_id)
        state_job = finder.latest(db.session)
        # Validate existence first as a job may not be queued yet as a result of
        # another prerequisite async process (dbt installation for example)
        if state_job:
            state_job_success = finder.latest_success(db.session)
            jobs.append(
                {
                    "state_id": state_id,
                    "is_complete": state_job.is_complete(),
                    "has_error": state_job.has_error(),
                    "started_at": state_job.started_at,
                    "ended_at": state_job.ended_at,
                    "has_ever_succeeded": state_job_success.is_success()
                    if state_job_success
                    else None,
                }
            )

    return jsonify({"jobs": jobs})


@orchestrations_bp.route("/jobs/<state_id>/log", methods=["GET"])
def job_log(state_id) -> Response:
    """Endpoint for getting the most recent log generated by a job with state_id.

    Args:
        state_id: id of the job you want to see logs of.

    Returns:
        JSON containing the jobs log entries
    """
    project = Project.find()
    try:
        log_service = JobLoggingService(project)
        log = log_service.get_latest_log(state_id)
        has_log_exceeded_max_size = False
    except SizeThresholdJobLogException:
        log = None
        has_log_exceeded_max_size = True

    finder = JobFinder(state_id)
    state_job = finder.latest(db.session)
    state_job_success = finder.latest_success(db.session)

    return jsonify(
        {
            "state_id": state_id,
            "log": log,
            "has_log_exceeded_max_size": has_log_exceeded_max_size,
            "has_error": state_job.has_error() if state_job else False,
            "started_at": state_job.started_at if state_job else None,
            "ended_at": state_job.ended_at if state_job else None,
            "trigger": state_job.trigger if state_job else None,
            "has_ever_succeeded": state_job_success.is_success()
            if state_job_success
            else None,
        }
    )


@orchestrations_bp.route("/jobs/<state_id>/download", methods=["GET"])
def download_job_log(state_id) -> Response:
    """Endpoint for downloading a job log with state_id.

    Args:
        state_id: id of the job you want log info of.

    Returns:
        A plain text file of the job log.
    """
    project = Project.find()
    log_service = JobLoggingService(project)
    return send_file(log_service.get_downloadable_log(state_id), mimetype="text/plain")


@orchestrations_bp.route("/run", methods=["POST"])
def run():
    """Run all scheduled payloads.

    Returns:
        JSON of all the state_ids that just ran.
    """
    project = Project.find()
    schedule_payload = request.get_json()
    name = schedule_payload["name"]
    state_id = run_schedule(project, name)

    return jsonify({"state_id": state_id}), 202


@orchestrations_bp.route(
    "/<plugin_ref:plugin_ref>/configuration/upload-file", methods=["POST"]
)
@block_if_readonly
def upload_plugin_configuration_file(plugin_ref) -> Response:
    """Endpoint for uploading a file for a specific plugin.

    Args:
        plugin_ref: Plugin being referenced.

    Returns:
        JSON contain the file path and setting name.
    """
    file = request.files["file"]
    setting_name = enforce_secure_filename(request.form["setting_name"])
    tmp = request.form.get("tmp", False)

    project = Project.find()
    directory = project.extract_dir(
        plugin_ref.name, ("tmp" if tmp else ""), setting_name
    )
    upload_helper = UploadHelper()
    file_path = upload_helper.upload_file(directory, file)

    return jsonify({"path": file_path, "setting_name": setting_name}), 200


@orchestrations_bp.route(
    "/<plugin_ref:plugin_ref>/configuration/delete-uploaded-file", methods=["POST"]
)
@block_if_readonly
def delete_plugin_configuration_file(plugin_ref) -> Response:
    """Endpoint for deleting a file for a specific plugin.

    Args:
        plugin_ref: Plugin being referenced.

    Returns:
        JSON contain the setting name.
    """
    payload = request.get_json()
    setting_name = enforce_secure_filename(payload["setting_name"])
    tmp = payload.get("tmp", False)

    project = Project.find()
    directory = project.extract_dir(
        plugin_ref.name, ("tmp" if tmp else ""), setting_name
    )
    shutil.rmtree(directory)

    return jsonify({"setting_name": setting_name}), 200


@orchestrations_bp.route("/<plugin_ref:plugin_ref>/configuration", methods=["GET"])
def get_plugin_configuration(plugin_ref) -> Response:
    """Endpoint for getting a plugin's configuration.

    Args:
        plugin_ref: Plugin being referenced.

    Returns:
        JSON contain the plugin configuration.
    """
    project = Project.find()

    plugins_service = ProjectPluginsService(project)
    plugin = plugins_service.get_plugin(plugin_ref)

    settings = PluginSettingsService(
        project, plugin, plugins_service=plugins_service, show_hidden=False
    )

    try:
        settings_group_validation = plugin.settings_group_validation
    except PluginNotFoundError:
        settings_group_validation = []

    return jsonify(
        {
            **get_config_with_metadata(settings),
            "settings": Canonical.as_canonical(settings.definitions(extras=False)),
            "settings_group_validation": settings_group_validation,
        }
    )


@orchestrations_bp.route("/<plugin_ref:plugin_ref>/configuration", methods=["PUT"])
@block_if_readonly
def save_plugin_configuration(plugin_ref) -> Response:
    """Endpoint for persisting a plugin configuration.

    Args:
        plugin_ref: Plugin being referenced.

    Returns:
        JSON contain the saved configuration.
    """
    project = Project.find()
    payload = request.get_json()
    plugins_service = ProjectPluginsService(project)
    plugin = plugins_service.get_plugin(plugin_ref)

    settings = PluginSettingsService(
        project, plugin, plugins_service=plugins_service, show_hidden=False
    )

    config = payload.get("config", {})
    for name, value in config.items():
        if not validate_plugin_config(plugin, name, value, project, settings):
            continue

        if value == "":
            settings.unset(name, session=db.session)
        else:
            settings.set(name, value, session=db.session)

    return jsonify(get_config_with_metadata(settings))


@orchestrations_bp.route(
    "/<plugin_ref:plugin_ref>/configuration/test", methods=["POST"]
)
@block_if_readonly
def test_plugin_configuration(plugin_ref) -> Response:  # noqa: WPS210
    """Endpoint for testing a plugin configuration's valid connection.

    Args:
        plugin_ref: Plugin being referenced.

    Returns:
        JSON with the job sucess status.
    """
    project = Project.find()
    plugins_service = ProjectPluginsService(project)
    plugin = plugins_service.get_plugin(plugin_ref)

    settings = PluginSettingsService(
        project, plugin, plugins_service=plugins_service, show_hidden=False
    )

    settings.config_override = PluginSettingsService.unredact(
        {
            name: value
            for name, value in request.get_json().get("config", {}).items()
            if validate_plugin_config(plugin, name, value, project, settings)
        }
    )

    async def test_extractor():
        invoker = invoker_factory(
            project,
            plugin,
            plugins_service=plugins_service,
            plugin_settings_service=settings,
        )
        async with invoker.prepared(db.session):
            plugin_test_service = PluginTestServiceFactory(invoker).get_test_service()
            success, _ = await plugin_test_service.validate()
            return success

    # This was added to assist api_worker threads
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        logging.debug("../configuration/test no asyncio event loop detected")
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    return jsonify({"is_success": loop.run_until_complete(test_extractor())}), 200


@orchestrations_bp.route("/pipeline-schedules", methods=["GET"])
def get_pipeline_schedules():
    """Endpoint for getting the pipeline schedules.

    Note that unless the ff ENABLE_API_SCHEDULED_JOB_LIST is enabled this endpoint will filter out scheduled jobs.

    Returns:
        JSON containing the pipline schedules.
    """
    project = Project.find()
    schedule_service = ScheduleService(project)
    schedules = list(map(dict, schedule_service.schedules()))

    jobs_in_list = False
    with ProjectSettingsService(project).feature_flag(
        FeatureFlags.ENABLE_API_SCHEDULED_JOB_LIST, raise_error=False
    ) as allow:
        if allow:
            jobs_in_list = True

    formatted_schedules = []

    for schedule in schedules:
        if schedule.get("job") and jobs_in_list:
            # we only return API results for scheduled jobs if the feature flag is explicitly enabled
            # as the UI is not job aware yet.
            formatted_schedules.append(schedule)
        elif not schedule.get("job"):  # a legacy elt task
            finder = JobFinder(schedule["name"])
            state_job = finder.latest(db.session)
            schedule["has_error"] = state_job.has_error() if state_job else False
            schedule["is_running"] = state_job.is_running() if state_job else False
            schedule["state_id"] = schedule["name"]
            schedule["started_at"] = state_job.started_at if state_job else None
            schedule["ended_at"] = state_job.ended_at if state_job else None
            schedule["trigger"] = state_job.trigger if state_job else None

            state_job_success = finder.latest_success(db.session)
            schedule["has_ever_succeeded"] = (
                state_job_success.is_success() if state_job_success else None
            )

            schedule["start_date"] = (  # noqa: WPS204
                schedule["start_date"].date().isoformat()
                if schedule["start_date"]
                else None
            )
            formatted_schedules.append(schedule)

    return jsonify(formatted_schedules)


@orchestrations_bp.route("/pipeline-schedules", methods=["POST"])
@block_if_readonly
def save_pipeline_schedule() -> Response:
    """Endpoint for persisting a pipeline schedule.

    Returns:
        JSON containing the saved pipline scheudles and status.
    """
    payload = request.get_json()
    # Airflow requires alphanumeric characters, dashes, dots and underscores exclusively
    name = payload["name"]
    slug = slugify(name)
    extractor = payload["extractor"]
    loader = payload["loader"]
    transform = payload["transform"]
    interval = payload["interval"]

    project = Project.find()
    schedule_service = ScheduleService(project)

    schedule = schedule_service.add_elt(
        db.session, slug, extractor, loader, transform, interval
    )

    schedule = dict(schedule)
    schedule["start_date"] = (
        schedule["start_date"].date().isoformat() if schedule["start_date"] else None
    )

    return jsonify(schedule), 201


@orchestrations_bp.route("/pipeline-schedules", methods=["PUT"])
@block_if_readonly
def update_pipeline_schedule() -> Response:
    """Endpoint for updating a pipeline schedule.

    Returns:
        JSON containing the updated pipeline schedules and status.
    """
    payload = request.get_json()
    project = Project.find()
    schedule_service = ScheduleService(project)

    plugin_namespace = payload["plugin_namespace"]
    schedule = schedule_service.find_namespace_schedule(plugin_namespace)

    if "interval" in payload:
        schedule.interval = payload.get("interval")

    if "transform" in payload:
        schedule.transform = payload.get("transform")

    schedule_service.update_schedule(schedule)

    schedule = dict(schedule)
    schedule["start_date"] = (
        schedule["start_date"].date().isoformat() if schedule["start_date"] else None
    )

    return jsonify(schedule), 201


@orchestrations_bp.route("/pipeline-schedules", methods=["DELETE"])
@block_if_readonly
def delete_pipeline_schedule() -> Response:
    """Endpoint for deleting a pipeline schedule.

    Returns:
        JSON containing the deleted pipelinea and status.
    """
    payload = request.get_json()
    project = Project.find()
    schedule_service = ScheduleService(project)

    name = payload["name"]
    schedule_service.remove(name)
    return jsonify(name), 201


class SubscriptionsResource(Resource):
    """Class that assists with managing resource subscriptions."""

    SubscriptionDefinition = {
        "id": fields.String,
        "recipient": fields.String,
        "event_type": fields.String,
        "source_type": fields.String,
        "source_id": fields.String,
        "created_at": fields.DateTime,
    }

    @marshal_with(SubscriptionDefinition)
    def get(self):
        """Get subscription.

        Returns:
            All subscription info.
        """
        return Subscription.query.all()

    @marshal_with(SubscriptionDefinition)
    def post(self):
        """Post subscriptions.

        Raises:
            Conflict: SQLAlchemy has detect a IntegirityError.
            UnprocessableEntity: if the SubscriptionDefintion is unprocessable.

        Returns:
            The subscription posted and status.
        """
        payload = request.get_json()

        try:
            subscription = Subscription(**payload)
            db.session.add(subscription)
            db.session.commit()
        except AssertionError as err:
            raise UnprocessableEntity() from err
        except sqlalchemy.exc.IntegrityError:
            raise Conflict()

        return subscription, 201


class SubscriptionResource(Resource):
    """Class that assists with managing resource subscriptions."""

    def delete(self, subscription_id):
        """Post subscriptions.

        Args:
            subscription_id: Subscription ID.

        Returns:
            The subscription delted and status.
        """
        Subscription.query.filter_by(id=subscription_id).delete()
        db.session.commit()

        return "", 204


orchestrationsAPI.add_resource(SubscriptionsResource, "/subscriptions")
orchestrationsAPI.add_resource(SubscriptionResource, "/subscriptions/<id>")



from __future__ import annotations

import fnmatch
import logging
import re
from collections import OrderedDict
from enum import Enum, auto
from functools import singledispatch
from typing import Any, Dict, Iterable, NamedTuple, TypeVar

from meltano.core.behavior.visitor import visit_with

Node = Dict[str, Any]
T = TypeVar("T", bound="CatalogRule")


class CatalogRule:
    def __init__(
        self,
        tap_stream_id: str | list[str],
        breadcrumb: list[str] | None = None,
        negated: bool = False,
    ):
        """Create a catalog rule for a stream and property."""
        self.tap_stream_id = tap_stream_id
        self.breadcrumb = breadcrumb or []
        self.negated = negated

    @classmethod
    def matching(
        cls: type[T],
        rules: list[T],
        tap_stream_id: str,
        breadcrumb: list[str] | None = None,
    ):
        """Filter rules that match a given breadcrumb."""
        return [rule for rule in rules if rule.match(tap_stream_id, breadcrumb)]

    def match(self, tap_stream_id: str, breadcrumb: list[str] | None = None) -> bool:
        """Evaluate if rule matches a stream or breadcrumb.

        Args:
            tap_stream_id: Singer stream identifier.
            breadcrumb: JSON property breadcrumb.

        Returns:
            A boolean representing whether the stream ID or breadcrumb matches the rules.
        """
        patterns = (
            self.tap_stream_id
            if isinstance(self.tap_stream_id, list)
            else [self.tap_stream_id]
        )

        result = any(fnmatch.fnmatch(tap_stream_id, pattern) for pattern in patterns)

        # A negated rule matches a stream ID when none of the patterns match
        if self.negated:
            result = not result

        # If provided, the breadcrumb should still match, even on negated rules
        if breadcrumb is not None:
            result = result and fnmatch.fnmatch(
                ".".join(breadcrumb), ".".join(self.breadcrumb)
            )

        return result


class MetadataRule(CatalogRule):
    def __init__(
        self,
        tap_stream_id: str | list[str],
        breadcrumb: list[str] | None,
        key: str,
        value: bool,
        negated: bool = False,
    ):
        """Create a metadata rule for a stream and property."""
        super().__init__(tap_stream_id, breadcrumb, negated=negated)
        self.key = key
        self.value = value


class SchemaRule(CatalogRule):
    def __init__(
        self,
        tap_stream_id: str | list[str],
        breadcrumb: list[str] | None,
        payload: dict,
        negated: bool = False,
    ):
        """Create a schema rule for a stream and property."""
        super().__init__(tap_stream_id, breadcrumb, negated=negated)
        self.payload = payload


class SelectPattern(NamedTuple):
    """A pattern for selecting streams and properties."""

    stream_pattern: str
    property_pattern: str | None
    negated: bool
    raw: str

    @classmethod
    def parse(cls, pattern: str):
        """Parse a SelectPattern instance from a string pattern.

        Args:
            pattern: Stream or property selection pattern.

        Returns:
            An appropriate `SelectPattern` instance.

        Example:

        >>> SelectPattern.parse("!a.b.c")
        SelectedPattern(stream_pattern='a', property_pattern='b.c', negated=True, raw='!a.b.c')
        """
        raw = pattern

        negated = False
        if pattern.startswith("!"):
            negated = True
            pattern = pattern[1:]

        if "." in pattern:
            stream, prop = pattern.split(".", maxsplit=1)
        else:
            stream = pattern
            prop = None

        return cls(
            stream_pattern=stream,
            property_pattern=prop,
            negated=negated,
            raw=raw,
        )


def select_metadata_rules(patterns: Iterable[str]) -> list[MetadataRule]:
    """Create metadata rules from `select` patterns.

    Args:
        patterns: Iterable of `select` string patterns.

    Returns:
        A list of corresponding metadata rule objects.
    """
    include_rules = []
    exclude_rules = []

    for pattern in patterns:
        pattern = SelectPattern.parse(pattern)

        prop_pattern = pattern.property_pattern
        selected = not pattern.negated

        rules = include_rules if selected else exclude_rules

        if selected or not prop_pattern or prop_pattern == "*":
            rules.append(
                MetadataRule(
                    tap_stream_id=pattern.stream_pattern,
                    breadcrumb=[],
                    key="selected",
                    value=selected,
                )
            )

        if prop_pattern:
            props = prop_pattern.split(".")

            rules.append(
                MetadataRule(
                    tap_stream_id=pattern.stream_pattern,
                    breadcrumb=property_breadcrumb(props),
                    key="selected",
                    value=selected,
                )
            )

    return include_rules + exclude_rules


def select_filter_metadata_rules(patterns: Iterable[str]) -> list[MetadataRule]:
    """Create metadata rules from `select_filter` patterns.

    Args:
        patterns: Iterable of `select_filter` string patterns.

    Returns:
        A list of corresponding metadata rule objects.
    """
    # We set `selected: false` if the `tap_stream_id`
    # does NOT match any of the selection/inclusion patterns
    include_rule = MetadataRule(
        negated=True, tap_stream_id=[], breadcrumb=[], key="selected", value=False
    )
    # Or if it matches one of the exclusion patterns
    exclude_rule = MetadataRule(
        tap_stream_id=[], breadcrumb=[], key="selected", value=False
    )

    for pattern in patterns:
        pattern = SelectPattern.parse(pattern)

        rule = exclude_rule if pattern.negated else include_rule
        rule.tap_stream_id.append(pattern.stream_pattern)

    rules = []
    if include_rule.tap_stream_id:
        rules.append(include_rule)
    if exclude_rule.tap_stream_id:
        rules.append(exclude_rule)

    return rules


def path_property(path: str) -> str:
    """Extract the property name from a materialized path.

    As we traverse the catalog tree, we build a materialized path
    to keep track of the parent nodes.

    Args:
        path: String representing a property path in the JSON schema.

    Returns:
        A string representing a property path in the JSON object.

    Examples:
      stream[0].properties.list_items.properties.account → list_items.account
      stream[0].properties.name                          → name
    """
    prop_regex = r"properties\.([\w\[\]\d]+)+"
    components = re.findall(prop_regex, path)
    return ".".join(components)


def property_breadcrumb(props: list[str]) -> list[str]:
    """Create breadcrumb from properties path list.

    Args:
        props: List of strings representing a property breadcrumb in the JSON object.

    Returns:
        A list of strings representing a property breadcrumb in the JSON schema.

    Example:
    >>> property_breadcrumb(["payload", "content"])
    ['properties', 'payload', 'properties', 'content']
    """
    if len(props) >= 2 and props[0] == "properties":
        breadcrumb = props
    else:
        breadcrumb = []
        for prop in props:
            breadcrumb.extend(["properties", prop])

    return breadcrumb


class CatalogNode(Enum):
    STREAM = auto()
    PROPERTY = auto()
    METADATA = auto()


class SelectionType(str, Enum):
    """A valid stream or property selection type."""

    SELECTED = "selected"
    EXCLUDED = "excluded"
    AUTOMATIC = "automatic"

    def __bool__(self):
        return self is not self.__class__.EXCLUDED

    def __add__(self, other):
        if self is SelectionType.EXCLUDED or other is SelectionType.EXCLUDED:
            return SelectionType.EXCLUDED

        if self is SelectionType.AUTOMATIC or other is SelectionType.AUTOMATIC:
            return SelectionType.AUTOMATIC

        return SelectionType.SELECTED


@singledispatch
def visit(node, executor, path: str = ""):
    logging.debug("Skipping node at '%s'", path)  # noqa: WPS323


@visit.register(dict)
def _(node: dict, executor, path=""):
    node_type = None

    if re.search(r"streams\[\d+\]$", path):
        node_type = CatalogNode.STREAM

    if re.search(r"schema(\.properties\.\w*)+$", path):
        node_type = CatalogNode.PROPERTY

    if re.search(r"metadata\[\d+\]$", path) and "breadcrumb" in node:
        node_type = CatalogNode.METADATA

    if node_type:
        logging.debug("Visiting %s at '%s'.", node_type, path)  # noqa: WPS323
        executor(node_type, node, path)

    for child_path, child_node in node.items():
        if node_type is CatalogNode.PROPERTY and child_path in {"anyOf", "type"}:
            continue

        # TODO mbergeron: refactor this to use a dynamic visitor per CatalogNode
        executor.visit(child_node, path=f"{path}.{child_path}")


@visit.register(list)
def _(node: list, executor, path=""):
    for index, child_node in enumerate(node):
        executor.visit(child_node, path=f"{path}[{index}]")


@visit_with(visit)
class CatalogExecutor:
    def execute(self, node_type: CatalogNode, node: Node, path: str):
        """Dispatch all node methods."""
        dispatch = {
            CatalogNode.STREAM: self.stream_node,
            CatalogNode.PROPERTY: self.property_node,
            CatalogNode.METADATA: self.metadata_node,
        }

        try:
            dispatch[node_type](node, path)
        except KeyError:
            logging.debug("Unknown node type '%s'.", node_type)  # noqa: WPS323

    def stream_node(self, node: Node, path: str):
        """Process stream node."""

    def property_node(self, node: Node, path: str):
        """Process property node."""

    def metadata_node(self, node: Node, path: str):
        """Process metadata node."""
        if len(node["breadcrumb"]) == 0:
            self.stream_metadata_node(node, path)
        else:
            self.property_metadata_node(node, path)

    def stream_metadata_node(self, node: Node, path: str):
        """Process stream metadata node."""

    def property_metadata_node(self, node: Node, path: str):
        """Process property metadata node."""

    def __call__(self, node_type, node: Node, path: str):
        """Call this instance as a function."""
        return self.execute(node_type, node, path)


class MetadataExecutor(CatalogExecutor):
    def __init__(self, rules: list[MetadataRule]):
        self._stream = None
        self._rules = rules

    def ensure_metadata(self, breadcrumb: list[str]):
        """Handle missing metadata entries."""
        metadata_list: list[dict] = self._stream["metadata"]
        match = next(
            (
                metadata
                for metadata in metadata_list
                if metadata["breadcrumb"] == breadcrumb
            ),
            None,
        )

        # Missing inclusion metadata for property
        if match is None:
            # Streams and top-level properties.
            if len(breadcrumb) <= 2:
                entry = {
                    "breadcrumb": breadcrumb,
                    "metadata": {"inclusion": "automatic"},
                }
            # Exclude nested properties.
            else:
                entry = {
                    "breadcrumb": breadcrumb,
                    "metadata": {"inclusion": "available"},
                }

            metadata_list.append(entry)

    def stream_node(self, node: Node, path: str):
        """Process stream metadata node."""
        self._stream = node
        tap_stream_id = self._stream["tap_stream_id"]

        if "metadata" not in node:
            node["metadata"] = []

        self.ensure_metadata([])

        for rule in MetadataRule.matching(self._rules, tap_stream_id, []):
            # Legacy catalogs have underscorized keys on the streams themselves
            self.set_metadata(node, path, rule.key.replace("-", "_"), rule.value)

    def property_node(self, node: Node, path: str):
        """Process property metadata node."""
        breadcrumb_idx = path.index("properties")
        breadcrumb = path[breadcrumb_idx:].split(".")

        self.ensure_metadata(breadcrumb)

    def metadata_node(self, node: Node, path: str):
        """Process metadata node."""
        tap_stream_id = self._stream["tap_stream_id"]
        breadcrumb = node["breadcrumb"]

        logging.debug(
            "Visiting metadata node for tap_stream_id '%s', breadcrumb '%s'",  # noqa: WPS323
            tap_stream_id,
            breadcrumb,
        )

        for rule in MetadataRule.matching(self._rules, tap_stream_id, breadcrumb):
            self.set_metadata(
                node["metadata"], f"{path}.metadata", rule.key, rule.value
            )

    def set_metadata(self, node: Node, path: str, key: str, value: Any):
        """Set selection and inclusion keys in a metadata node."""
        # Unsupported fields cannot be selected
        if (
            key == "selected"
            and value is True
            and node.get("inclusion") == "unsupported"
        ):
            return

        node[key] = value
        logging.debug("Setting '%s.%s' to '%s'", path, key, value)  # noqa: WPS323


class SelectExecutor(MetadataExecutor):
    def __init__(self, patterns: list[str]):
        super().__init__(select_metadata_rules(patterns))


class SchemaExecutor(CatalogExecutor):
    def __init__(self, rules: list[SchemaRule]):
        self._stream = None
        self._rules = rules

    def ensure_property(self, breadcrumb: list[str]):  # noqa: WPS231
        """Create nodes for the breadcrumb and schema extra that matches."""
        next_node: dict[str, Any] = self._stream["schema"]

        for idx, key in enumerate(breadcrumb):
            # If the key contains shell-style wildcards,
            # ensure property nodes exist for matching breadcrumbs.
            if re.match(r"[*?\[\]]", key):
                node_keys = next_node.keys()
                matching_keys = fnmatch.filter(node_keys, key)

                if matching_keys:
                    matching_breadcrumb = breadcrumb.copy()
                    for key in matching_keys:
                        matching_breadcrumb[idx] = key
                        self.ensure_property(matching_breadcrumb)

                break

            # If a property node for this breadcrumb doesn't exist yet, create it.
            if key not in next_node:
                next_node[key] = {}

            next_node = next_node[key]

    def stream_node(self, node: Node, path):
        """Process stream schema node."""
        self._stream = node
        tap_stream_id: str = self._stream["tap_stream_id"]

        if "schema" not in node:
            node["schema"] = {"type": "object"}

        for rule in SchemaRule.matching(self._rules, tap_stream_id):
            self.ensure_property(rule.breadcrumb)

    def property_node(self, node: Node, path: str):
        """Process property schema node."""
        tap_stream_id = self._stream["tap_stream_id"]

        breadcrumb_idx = path.index("properties")
        breadcrumb = path[breadcrumb_idx:].split(".")

        for rule in SchemaRule.matching(self._rules, tap_stream_id, breadcrumb):
            self.set_payload(node, path, rule.payload)

    def set_payload(self, node: Node, path: str, payload: dict):
        """Set node payload from a clean mapping."""
        node.clear()
        node.update(payload)
        logging.debug("Setting '%s' to %r", path, payload)  # noqa: WPS323


class ListExecutor(CatalogExecutor):
    def __init__(self):
        # properties per stream
        self.properties: dict[str, set[str]] = OrderedDict()

        super().__init__()

    def stream_node(self, node: Node, path: str):
        """Initialize empty property set stream."""
        stream = node["tap_stream_id"]
        if stream not in self.properties:
            self.properties[stream] = set()

    def property_node(self, node: Node, path: str):
        """Add property to stream collection."""
        prop = path_property(path)
        # current stream
        stream = next(reversed(self.properties))
        self.properties[stream].add(prop)


class SelectedNode(NamedTuple):
    """Selection type and key of a node."""

    key: str
    selection: SelectionType


class ListSelectedExecutor(CatalogExecutor):
    def __init__(self):
        self.streams: set[SelectedNode] = set()
        self.properties: dict[str, set[SelectedNode]] = OrderedDict()
        super().__init__()

    @property
    def selected_properties(self):
        """Get selected streams and properties."""
        # we don't want to mutate the visitor result
        selected_properties = self.properties.copy()

        # remove all non-selected streams
        for name, selected in self.streams:
            if not selected:
                del selected_properties[name]

        # remove all non-selected properties
        for stream, props in selected_properties.items():
            selected_properties[stream] = {name for name, selected in props if selected}

        return selected_properties

    @staticmethod
    def node_selection(node: Node) -> SelectionType:
        """Get selection type from metadata entry.

        Args:
            node: Catalog metadata dictionary.

        Returns:
            A proper `SelectionType` given the inclusion and selection metadata.
        """
        try:
            metadata: dict[str, Any] = node["metadata"]
        except KeyError:
            return SelectionType.EXCLUDED

        inclusion: str = metadata.get("inclusion")
        selected: bool | None = metadata.get("selected")
        selected_by_default: bool = metadata.get("selected-by-default", False)

        if inclusion == "automatic":
            return SelectionType.AUTOMATIC

        if selected is True:
            return SelectionType.SELECTED

        if selected is None and selected_by_default:
            return SelectionType.SELECTED

        return SelectionType.EXCLUDED

    def stream_node(self, node: Node, path: str):
        """Initialize empty set for selected nodes in stream."""
        self._stream: str = node["tap_stream_id"]
        self.properties[self._stream] = set()

    def stream_metadata_node(self, node: Node, path: str):
        """Add stream selection to tap's collection."""
        selection = SelectedNode(self._stream, self.node_selection(node))
        self.streams.add(selection)

    def property_metadata_node(self, node: Node, path: str):
        """Add property selection to stream's collection."""
        property_path = ".".join(node["breadcrumb"])
        prop = path_property(property_path)
        selection = SelectedNode(prop, self.node_selection(node))

        self.properties[self._stream].add(selection)



from __future__ import annotations

import json

from meltano.core.plugin import PluginType
from meltano.core.plugin.base import PluginRef
from meltano.core.plugin.error import PluginExecutionError
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.plugin.singer.catalog import ListSelectedExecutor
from meltano.core.plugin_invoker import invoker_factory
from meltano.core.project_plugins_service import ProjectPluginsService

from .project import Project


class SelectService:
    def __init__(
        self,
        project: Project,
        extractor: str,
        plugins_service: ProjectPluginsService | None = None,
    ):
        self.project = project
        self.plugins_service = plugins_service or ProjectPluginsService(project)
        self._extractor = self.plugins_service.find_plugin(
            extractor, PluginType.EXTRACTORS
        )

    @property
    def extractor(self) -> ProjectPlugin:
        """Retrieve extractor ProjectPlugin object."""
        return self._extractor

    @property
    def current_select(self):
        plugin_settings_service = PluginSettingsService(
            self.project,
            self.extractor,
            plugins_service=self.plugins_service,
        )
        return plugin_settings_service.get("_select")

    async def load_catalog(self, session):
        """Load the catalog."""
        invoker = invoker_factory(
            self.project,
            self.extractor,
            plugins_service=self.plugins_service,
        )

        async with invoker.prepared(session):
            catalog_json = await invoker.dump("catalog")

        return json.loads(catalog_json)

    async def list_all(self, session) -> ListSelectedExecutor:
        """List all select."""
        try:
            catalog = await self.load_catalog(session)
        except FileNotFoundError as err:
            raise PluginExecutionError(
                "Could not find catalog. Verify that the tap supports discovery "
                + "mode and advertises the `discover` capability as well as either "
                + "`catalog` or `properties`"
            ) from err

        list_all = ListSelectedExecutor()
        list_all.visit(catalog)

        return list_all

    def update(self, entities_filter, attributes_filter, exclude, remove=False):
        """Update plugins' select patterns."""
        plugin: PluginRef

        if self.project.active_environment is None:
            plugin = self.extractor
        else:
            plugin = self.project.active_environment.get_plugin_config(
                self.extractor.type, self.extractor.name
            )

        this_pattern = self._get_pattern_string(
            entities_filter, attributes_filter, exclude
        )
        patterns = plugin.extras.get("select", [])
        if remove:
            patterns.remove(this_pattern)
        else:
            patterns.append(this_pattern)
        plugin.extras["select"] = patterns

        if self.project.active_environment is None:
            self.plugins_service.update_plugin(plugin)
        else:
            self.plugins_service.update_environment_plugin(plugin)

    @staticmethod
    def _get_pattern_string(entities_filter, attributes_filter, exclude) -> str:
        """Return a select pattern in string form."""
        exclude = "!" if exclude else ""
        return f"{exclude}{entities_filter}.{attributes_filter}"



"""Defines JobState model class."""
from __future__ import annotations

import json
from datetime import datetime
from io import TextIOWrapper
from typing import Any

from sqlalchemy import Column, types
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import Mapped, Session

from meltano.core.job import JobFinder, Payload
from meltano.core.models import SystemModel
from meltano.core.sqlalchemy import JSONEncodedDict
from meltano.core.utils import merge

SINGER_STATE_KEY = "singer_state"
STATE_ID_COMPONENT_DELIMITER = ":"


class JobState(SystemModel):  # noqa: WPS214
    """Model class that represents the current state of a given job.

    Modified during `meltano elt` or `meltano run` runs whenever a
    STATE message is emitted by a Singer target. Also written and read
    by `meltano state` CLI invocations. Only holds the _current_ state
    for a given state_id. Full job run history is held by the Job model.
    """

    __tablename__ = "state"
    state_id = Column(types.String, unique=True, primary_key=True, nullable=False)

    updated_at = Column(types.DATETIME, onupdate=datetime.now)

    partial_state: Mapped[Any] = Column(MutableDict.as_mutable(JSONEncodedDict))
    completed_state: Mapped[Any] = Column(MutableDict.as_mutable(JSONEncodedDict))

    def __eq__(self, other: object) -> bool:
        """Check equality with another JobState.

        Does not take into account updated_at;
        merely checks equality of state_id and partial and completed state.

        Args:
            other: the other JobState to check against

        Returns:
            True if other is equal to this JobState, else False
        """
        if not isinstance(other, JobState):
            return NotImplemented
        return (
            (self.state_id == other.state_id)
            and (self.partial_state == other.partial_state)
            and (self.completed_state == other.completed_state)
        )

    @classmethod  # noqa: WPS210
    def from_job_history(cls, session: Session, state_id: str):  # noqa: WPS210
        """Build JobState from job run history.

        Args:
            session: the session to use in finding job history
            state_id: state_id to build JobState for

        Returns:
            JobState built from job run history
        """
        completed_state: dict[Any, Any] = {}
        partial_state: dict[Any, Any] = {}
        incomplete_since = None
        finder = JobFinder(state_id)

        # Get the state for the most recent completed job.
        # Do not consider dummy jobs create via add_state.
        state_job = finder.latest_with_payload(session, flags=Payload.STATE)
        if state_job:
            incomplete_since = state_job.ended_at
            if SINGER_STATE_KEY in state_job.payload:
                merge(state_job.payload, partial_state)

        # If there have been any incomplete jobs since the most recent completed jobs,
        # merge the state emitted by those jobs into the state for the most recent
        # completed job. If there are no completed jobs, get the full history of
        # incomplete jobs and use the most recent state emitted per stream
        incomplete_state_jobs = finder.with_payload(
            session, flags=Payload.INCOMPLETE_STATE, since=incomplete_since
        )
        for incomplete_state_job in incomplete_state_jobs:
            if SINGER_STATE_KEY in incomplete_state_job.payload:
                partial_state = merge(incomplete_state_job.payload, partial_state)

        return cls(
            state_id=state_id,
            partial_state=partial_state,
            completed_state=completed_state,
        )

    @classmethod
    def from_json(cls, state_id: str, json_str: str) -> JobState:
        """Create JobState from json representation.

        Args:
            state_id: the state ID of the job to create JobState for.
            json_str: the json representation of state to use.

        Returns:
            JobState representing args.
        """
        state_dict = json.loads(json_str)
        return cls(
            state_id=state_id,
            completed_state=state_dict.get("completed", {}),
            partial_state=state_dict.get("partial", {}),
        )

    def json(self) -> str:
        """Get the json representation of this JobState.

        Returns:
            json representation of this JobState
        """
        return json.dumps(
            {"completed": self.completed_state, "partial": self.partial_state}
        )

    def json_merged(self) -> str:
        """Return the json representation of partial state merged onto complete state.

        Returns:
            json representation of partial state merged onto complete state.
        """
        return json.dumps(merge(self.partial_state, self.completed_state))

    @classmethod
    def from_file(cls, state_id: str, file_obj: TextIOWrapper) -> JobState:
        """Create JobState from a file-like object containing state json.

        Args:
            state_id: the state_id for the JobState
            file_obj: the file-like object containing state json.

        Returns:
            JobState
        """
        return cls.from_json(state_id=state_id, json_str=file_obj.read())

    def merge_partial(
        self,
        state: JobState,
    ):
        """Merge provided partial state onto this JobState.

        Args:
            state: the partial state to merge onto this state.
        """
        self.partial_state = merge(  # noqa: WPS601
            state.partial_state, self.partial_state
        )

    def is_complete(self) -> bool:
        """Check if this JobState is complete.

        Returns:
            True if complete, else False
        """
        return bool(self.completed_state)

    def to_file(self, file_obj: TextIOWrapper):
        """Persist JobState to a file-like object.

        Args:
            file_obj: the file-like object to write to.
        """
        file_obj.write(self.json())



"""The Meltano REPL."""

from __future__ import annotations

import click

from meltano.cli.cli import cli
from meltano.cli.params import database_uri_option
from meltano.cli.utils import InstrumentedCmd


@cli.command(cls=InstrumentedCmd, hidden=True)
@database_uri_option
@click.pass_context
def repl(ctx: click.Context):
    """Start the Meltano REPL."""
    try:
        import IPython
    except ImportError as ex:
        click.secho("The 'ipython' package must be installed to use the REPL", fg="red")
        raise click.Abort from ex

    from traitlets.config import Config

    # First create a config object from the traitlets library
    config = Config()

    config.InteractiveShellApp.extensions = ["autoreload"]
    config.InteractiveShellApp.exec_lines = [
        r'print("\nBooting import Meltano REPL\n")',
        "from meltano.core.project import Project",
        "from meltano.core.project_settings_service import ProjectSettingsService",
        "from meltano.core.project_plugins_service import ProjectPluginsService",
        "from meltano.core.plugin.settings_service import PluginSettingsService",
        "from meltano.core.db import project_engine",
        "project = Project.find()",
        "settings_service = ProjectSettingsService(project)",
        "_, Session = project_engine(project, default=True)",
        "session = Session()",
        "%autoreload 2",  # noqa: WPS323
    ]
    # config.InteractiveShell.colors = 'LightBG'  # noqa: E800
    config.InteractiveShell.confirm_exit = False
    config.TerminalIPythonApp.display_banner = True

    # Now we start ipython with our configuration
    IPython.start_ipython(argv=[], config=config)



"""Create meltano.core base tables

Revision ID: b4c05e463b53
Revises: 6f28844bcd3c
Create Date: 2019-07-23 16:05:29.073296

"""
from enum import Enum

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects.mssql import DATETIME2
from sqlalchemy.ext.mutable import MutableDict

from meltano.migrations import IntFlag, JSONEncodedDict
from meltano.migrations.utils.dialect_typing import (
    datetime_for_dialect,
    get_dialect_name,
    max_string_length_for_dialect,
)

# revision identifiers, used by Alembic.
revision = "b4c05e463b53"
down_revision = None
branch_labels = None
depends_on = None


# from `src/meltano/core/job.py`
class State(Enum):
    IDLE = (0, ("RUNNING", "FAIL"))
    RUNNING = (1, ("SUCCESS", "FAIL"))
    SUCCESS = (2, ())
    FAIL = (3, ("RUNNING",))
    DEAD = (4, ())


def upgrade():
    dialect_name = get_dialect_name()
    datetime_type = datetime_for_dialect(dialect_name)
    max_string_length = max_string_length_for_dialect(dialect_name)

    op.create_table(
        "job",
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column("job_id", sa.String(max_string_length)),
        sa.Column("state", sa.Enum(State, name="job_state")),
        sa.Column("started_at", datetime_type),
        sa.Column("ended_at", datetime_type),
        sa.Column(
            "payload", MutableDict.as_mutable(JSONEncodedDict(max_string_length))
        ),
        sa.Column("payload_flags", IntFlag, default=0),
    )

    op.create_table(
        "plugin_settings",
        sa.Column("label", datetime_type, nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("name", sa.String(255), nullable=False),
        sa.Column("namespace", sa.String(255), nullable=True),
        sa.Column("value", sa.PickleType(), nullable=True),
        sa.Column("enabled", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("name", "namespace"),
    )


def downgrade():
    op.drop_table("job")
    op.drop_table("plugin_settings")
